"""Request Extractors"""
from scrapy.http import Request
from scrapy.selector import HtmlXPathSelector
from scrapy.utils.misc import arg_to_iter
from scrapy.utils.python import FixedSGMLParser, str_to_unicode
from scrapy.utils.url import safe_url_string, urljoin_rfc

from itertools import ifilter


class BaseSgmlRequestExtractor(FixedSGMLParser):
    """Base SGML Request Extractor"""

    def __init__(self, tag='a', attr='href'):
        """Initialize attributes"""
        FixedSGMLParser.__init__(self)

        self.scan_tag = tag if callable(tag) else lambda t: t == tag
        self.scan_attr = attr if callable(attr) else lambda a: a == attr
        self.current_request = None

    def extract_requests(self, response):
        """Returns list of requests extracted from response"""
        return self._extract_requests(response.body, response.url,
                                  response.encoding)

    def _extract_requests(self, response_text, response_url, response_encoding):
        """Extract requests with absolute urls"""
        self.reset()
        self.feed(response_text)
        self.close()

        base_url = urljoin_rfc(response_url, self.base_url) if self.base_url else response_url
        self._make_absolute_urls(base_url, response_encoding)
        self._fix_link_text_encoding(response_encoding)

        return self.requests

    def _make_absolute_urls(self, base_url, encoding):
        """Makes all request's urls absolute"""
        self.requests = [x.replace(url=safe_url_string(urljoin_rfc(base_url, \
            x.url, encoding), encoding)) for x in self.requests]

    def _fix_link_text_encoding(self, encoding):
        """Convert link_text to unicode for each request"""
        for req in self.requests:
            req.meta.setdefault('link_text', '')
            req.meta['link_text'] = str_to_unicode(req.meta['link_text'],
                                                   encoding) 

    def reset(self):
        """Reset state"""
        FixedSGMLParser.reset(self)
        self.requests = []
        self.base_url = None
            
    def unknown_starttag(self, tag, attrs):
        """Process unknown start tag"""
        if 'base' == tag:
            self.base_url = dict(attrs).get('href')

        _matches = lambda (attr, value): self.scan_attr(attr) \
                                        and value is not None
        if self.scan_tag(tag):
            for attr, value in ifilter(_matches, attrs):
                req = Request(url=value)
                self.requests.append(req)
                self.current_request = req

    def unknown_endtag(self, tag):
        """Process unknown end tag"""
        self.current_request = None

    def handle_data(self, data):
        """Process data"""
        current = self.current_request
        if current and not 'link_text' in current.meta:
            current.meta['link_text'] = data.strip()


class SgmlRequestExtractor(BaseSgmlRequestExtractor):
    """SGML Request Extractor"""

    def __init__(self, tags=None, attrs=None):
        """Initialize with custom tag & attribute function checkers"""
        # defaults
        tags = tuple(tags) if tags else ('a', 'area')
        attrs = tuple(attrs) if attrs else ('href', )

        tag_func = lambda x: x in tags
        attr_func = lambda x: x in attrs
        BaseSgmlRequestExtractor.__init__(self, tag=tag_func, attr=attr_func)

# TODO: move to own file
class XPathRequestExtractor(SgmlRequestExtractor):
    """SGML Request Extractor with XPath restriction"""

    def __init__(self, restrict_xpaths, tags=None, attrs=None):
        """Initialize XPath restrictions"""
        self.restrict_xpaths = tuple(arg_to_iter(restrict_xpaths))
        SgmlRequestExtractor.__init__(self, tags, attrs)

    def extract_requests(self, response):
        """Restrict to XPath regions"""
        hxs = HtmlXPathSelector(response)
        fragments = (''.join(
                            html_frag for html_frag in hxs.select(xpath).extract()
                        ) for xpath in self.restrict_xpaths)
        html_slice = ''.join(html_frag for html_frag in fragments)
        return self._extract_requests(html_slice, response.url,
                                        response.encoding)




import os

from twisted.trial import unittest

from zope.interface.verify import verifyObject

from scrapyd.interfaces import IEnvironment
from scrapyd.config import Config
from scrapyd.environ import Environment

class EggStorageTest(unittest.TestCase):

    def setUp(self):
        d = self.mktemp()
        os.mkdir(d)
        config = Config(values={'eggs_dir': d, 'logs_dir': d})
        config.cp.add_section('settings')
        config.cp.set('settings', 'newbot', 'newbot.settings')
        self.environ = Environment(config)

    def test_interface(self):
        verifyObject(IEnvironment, self.environ)

    def test_get_environment_with_eggfile(self):
        msg = {'project': 'mybot'}
        slot = 3
        env = self.environ.get_environment(msg, slot, '/path/to/file.egg')
        self.assertEqual(env['SCRAPY_PROJECT'], 'mybot')
        self.assert_(env['SCRAPY_SQLITE_DB'].endswith('mybot.db'))
        self.assert_(env['SCRAPY_LOG_FILE'].endswith('slot3.log'))
        self.assert_(env['SCRAPY_EGGFILE'].endswith('/path/to/file.egg'))
        self.failIf('SCRAPY_SETTINGS_MODULE' in env)

    def test_get_environment_without_eggfile(self):
        msg = {'project': 'newbot'}
        slot = 3
        env = self.environ.get_environment(msg, slot, None)
        self.assertEqual(env['SCRAPY_PROJECT'], 'newbot')
        self.assert_(env['SCRAPY_SQLITE_DB'].endswith('newbot.db'))
        self.assert_(env['SCRAPY_LOG_FILE'].endswith('slot3.log'))
        self.assertEqual(env['SCRAPY_SETTINGS_MODULE'], 'newbot.settings')
        self.failIf('SCRAPY_EGGFILE' in env)



"""
Download web pages using asynchronous IO
"""

import random
from time import time

from twisted.internet import reactor, defer
from twisted.python.failure import Failure

from scrapy.exceptions import IgnoreRequest
from scrapy.conf import settings
from scrapy.utils.defer import mustbe_deferred
from scrapy.utils.signal import send_catch_log
from scrapy import signals
from scrapy import log
from .middleware import DownloaderMiddlewareManager
from .handlers import DownloadHandlers


class SpiderInfo(object):
    """Simple class to keep information and state for each open spider"""

    def __init__(self, download_delay=None, max_concurrent_requests=None):
        if download_delay is None:
            self._download_delay = settings.getfloat('DOWNLOAD_DELAY')
        else:
            self._download_delay = float(download_delay)
        if self._download_delay:
            self.max_concurrent_requests = 1
        elif max_concurrent_requests is None:
            self.max_concurrent_requests = settings.getint('CONCURRENT_REQUESTS_PER_SPIDER')
        else:
            self.max_concurrent_requests =  max_concurrent_requests
        if self._download_delay and settings.getbool('RANDOMIZE_DOWNLOAD_DELAY'):
            # same policy as wget --random-wait
            self.random_delay_interval = (0.5*self._download_delay, \
                1.5*self._download_delay)
        else:
            self.random_delay_interval = None

        self.active = set()
        self.queue = []
        self.transferring = set()
        self.closing = False
        self.lastseen = 0
        self.next_request_calls = set()

    def free_transfer_slots(self):
        return self.max_concurrent_requests - len(self.transferring)

    def needs_backout(self):
        # use self.active to include requests in the downloader middleware
        return len(self.active) > 2 * self.max_concurrent_requests

    def download_delay(self):
        if self.random_delay_interval:
            return random.uniform(*self.random_delay_interval)
        else:
            return self._download_delay

    def cancel_request_calls(self):
        for call in self.next_request_calls:
            call.cancel()
        self.next_request_calls.clear()


class Downloader(object):
    """Mantain many concurrent downloads and provide an HTTP abstraction.
    It supports a limited number of connections per spider and many spiders in
    parallel.
    """

    def __init__(self):
        self.sites = {}
        self.handlers = DownloadHandlers()
        self.middleware = DownloaderMiddlewareManager.from_settings(settings)
        self.concurrent_spiders = settings.getint('CONCURRENT_SPIDERS')

    def fetch(self, request, spider):
        """Main method to use to request a download

        This method includes middleware mangling. Middleware can returns a
        Response object, then request never reach downloader queue, and it will
        not be downloaded from site.
        """
        site = self.sites[spider]
        if site.closing:
            raise IgnoreRequest('Cannot fetch on a closing spider')

        site.active.add(request)
        def _deactivate(response):
            send_catch_log(signal=signals.response_received, \
                response=response, request=request, spider=spider)
            site.active.remove(request)
            self._close_if_idle(spider)
            return response

        dfd = self.middleware.download(self.enqueue, request, spider)
        return dfd.addBoth(_deactivate)

    def enqueue(self, request, spider):
        """Enqueue a Request for a effective download from site"""
        site = self.sites[spider]
        if site.closing:
            raise IgnoreRequest

        def _downloaded(response):
            send_catch_log(signal=signals.response_downloaded, \
                    response=response, request=request, spider=spider)
            return response

        deferred = defer.Deferred().addCallback(_downloaded)
        site.queue.append((request, deferred))
        self._process_queue(spider)
        return deferred

    def _process_queue(self, spider):
        """Effective download requests from site queue"""
        site = self.sites.get(spider)
        if not site:
            return

        # Delay queue processing if a download_delay is configured
        now = time()
        delay = site.download_delay()
        if delay:
            penalty = delay - now + site.lastseen
            if penalty > 0:
                d = defer.Deferred()
                d.addCallback(self._process_queue)
                call = reactor.callLater(penalty, d.callback, spider)
                site.next_request_calls.add(call)
                d.addBoth(lambda x: site.next_request_calls.remove(call))
                return
        site.lastseen = now

        # Process enqueued requests if there are free slots to transfer for this site
        while site.queue and site.free_transfer_slots() > 0:
            request, deferred = site.queue.pop(0)
            if site.closing:
                dfd = defer.fail(Failure(IgnoreRequest()))
            else:
                dfd = self._download(site, request, spider)
            dfd.chainDeferred(deferred)

        self._close_if_idle(spider)

    def _close_if_idle(self, spider):
        site = self.sites.get(spider)
        if site and site.closing and not site.active:
            del self.sites[spider]
            site.closing.callback(None)

    def _download(self, site, request, spider):
        # The order is very important for the following deferreds. Do not change!

        # 1. Create the download deferred
        dfd = mustbe_deferred(self.handlers.download_request, request, spider)

        # 2. After response arrives,  remove the request from transferring
        # state to free up the transferring slot so it can be used by the
        # following requests (perhaps those which came from the downloader
        # middleware itself)
        site.transferring.add(request)
        def finish_transferring(_):
            site.transferring.remove(request)
            self._process_queue(spider)
            # avoid partially downloaded responses from propagating to the
            # downloader middleware, to speed-up the closing process
            if site.closing:
                log.msg("Crawled while closing spider: %s" % request, \
                    level=log.DEBUG, spider=spider)
                raise IgnoreRequest
            return _
        return dfd.addBoth(finish_transferring)

    def open_spider(self, spider):
        """Allocate resources to begin processing a spider"""
        assert spider not in self.sites, "Spider already opened: %s" % spider
        self.sites[spider] = SpiderInfo(
            download_delay=getattr(spider, 'download_delay', None),
            max_concurrent_requests=getattr(spider, 'max_concurrent_requests', None)
        )

    def close_spider(self, spider):
        """Free any resources associated with the given spider"""
        assert spider in self.sites, "Spider not opened: %s" % spider
        site = self.sites.get(spider)
        site.closing = defer.Deferred()
        site.cancel_request_calls()
        self._process_queue(spider)
        return site.closing

    def is_idle(self):
        return not self.sites




import unittest
from datetime import datetime

from scrapy.utils.simpledb import to_sdb_value

class SimpleddbUtilsTest(unittest.TestCase):

    def test_to_sdb_value(self):
        self.assertEqual(to_sdb_value(123), u'0000000000000123')
        self.assertEqual(to_sdb_value(123L), u'0000000000000123')
        self.assertEqual(to_sdb_value(True), u'1')
        self.assertEqual(to_sdb_value(False), u'0')
        self.assertEqual(to_sdb_value(None), u'')
        self.assertEqual(to_sdb_value(datetime(2009, 01, 01, 10, 10, 10)), \
            u'2009-01-01T10:10:10')
        self.assertEqual(to_sdb_value('test'), 'test')
        self.assertEqual(to_sdb_value(u'test'), u'test')
        self.assertRaises(TypeError, to_sdb_value, object())

if __name__ == "__main__":
    unittest.main()



from zope.interface import implements

from twisted.internet import threads
from boto.sqs.connection import SQSConnection
from boto.sqs.message import Message
from boto.sqs import regions

from scrapy.interfaces import ISpiderQueue
from scrapy.utils.py26 import json

class SQSSpiderQueue(object):

    implements(ISpiderQueue)

    def __init__(self, *a, **kw):
        self.queue_name = kw.pop('queue_name', 'scrapy')
        self.region_name = kw.pop('region_name', 'us-east-1')
        self.visibility_timeout = kw.pop('visibility_timeout', 7200)
        self.aws_access_key_id = kw.pop('aws_access_key_id', None)
        self.aws_secret_access_key = kw.pop('aws_secret_access_key', None)
        self.region = self._get_region(self.region_name)
        self.conn.create_queue(self.queue_name)
        super(SQSSpiderQueue, self).__init__(*a, **kw)

    @classmethod
    def from_settings(cls, settings):
        return cls(
            queue_name=settings['SQS_QUEUE'],
            region_name=settings['SQS_REGION'],
            visibility_timeout=settings.getint('SQS_VISIBILITY_TIMEOUT'),
            aws_access_key_id=settings['AWS_ACCESS_KEY_ID'],
            aws_secret_access_key=settings['AWS_SECRET_ACCESS_KEY']
        )
        
    def _get_region(self, name):
        return [r for r in regions() if r.name == name][0]

    @property
    def conn(self):
        return SQSConnection(self.aws_access_key_id, self.aws_secret_access_key, \
            region=self.region)

    @property
    def queue(self):
        return self.conn.get_queue(self.queue_name)

    def _queue_method(self, method, *a, **kw):
        return getattr(self.queue, method)(*a, **kw)

    def pop(self):
        return threads.deferToThread(self._pop)

    def _pop(self):
        msgs = self.queue.get_messages(1, visibility_timeout=self.visibility_timeout)
        if msgs:
            msg = msgs[0]
            msg.delete()
            return json.loads(msg.get_body())

    def add(self, name, **spider_args):
        d = spider_args.copy()
        d['name'] = name
        msg = Message(body=json.dumps(d))
        return threads.deferToThread(self._queue_method, 'write', msg)

    def count(self):
        return threads.deferToThread(self._queue_method, 'count')

    def list(self):
        return threads.deferToThread(self._list)

    def _list(self):
        msgs = []
        m = self.queue.read(visibility_timeout=100)
        while m:
            msgs.append(json.loads(m.get_body()))
            m = self.queue.read(visibility_timeout=100)
        return msgs

    def clear(self):
        return threads.deferToThread(self._queue_method, 'clear')



"""
Scrapy extension for collecting scraping stats
"""
import os
import getpass
import socket
import datetime

from scrapy.xlib.pydispatch import dispatcher

from scrapy import signals
from scrapy.stats import stats
from scrapy.conf import settings

class CoreStats(object):
    """Scrapy core stats collector"""

    def __init__(self):
        stats.set_value('envinfo/user', getpass.getuser())
        stats.set_value('envinfo/host', socket.gethostname())
        stats.set_value('envinfo/logfile', settings['LOG_FILE'])
        stats.set_value('envinfo/pid', os.getpid())

        dispatcher.connect(self.stats_spider_opened, signal=signals.stats_spider_opened)
        dispatcher.connect(self.stats_spider_closing, signal=signals.stats_spider_closing)
        dispatcher.connect(self.item_scraped, signal=signals.item_scraped)
        dispatcher.connect(self.item_passed, signal=signals.item_passed)
        dispatcher.connect(self.item_dropped, signal=signals.item_dropped)

    def stats_spider_opened(self, spider):
        stats.set_value('start_time', datetime.datetime.utcnow(), spider=spider)
        stats.set_value('envinfo/host', stats.get_value('envinfo/host'), spider=spider)
        stats.inc_value('spider_count/opened')

    def stats_spider_closing(self, spider, reason):
        stats.set_value('finish_time', datetime.datetime.utcnow(), spider=spider)
        stats.set_value('finish_status', 'OK' if reason == 'finished' else reason, spider=spider)
        stats.inc_value('spider_count/%s' % reason, spider=spider)

    def item_scraped(self, item, spider):
        stats.inc_value('item_scraped_count', spider=spider)
        stats.inc_value('item_scraped_count')

    def item_passed(self, item, spider):
        stats.inc_value('item_passed_count', spider=spider)
        stats.inc_value('item_passed_count')

    def item_dropped(self, item, spider, exception):
        reason = exception.__class__.__name__
        stats.inc_value('item_dropped_count', spider=spider)
        stats.inc_value('item_dropped_reasons_count/%s' % reason, spider=spider)
        stats.inc_value('item_dropped_count')



"""
Page parsing

Parsing of web pages for extraction task.
"""
from collections import defaultdict
from numpy import array

from scrapy.utils.py26 import json

from scrapy.contrib.ibl.htmlpage import HtmlTagType, HtmlTag, HtmlPage
from scrapy.contrib.ibl.extraction.pageobjects import (AnnotationTag, 
        TemplatePage, ExtractionPage, AnnotationText, TokenDict)

def parse_strings(template_html, extraction_html):
    """Create a template and extraction page from raw strings

    this is useful for testing purposes
    """
    t = TokenDict()
    template_page = HtmlPage(body=template_html)
    extraction_page = HtmlPage(body=extraction_html)
    return (parse_template(t, template_page), 
            parse_extraction_page(t, extraction_page))

def parse_template(token_dict, template_html):
    """Create an TemplatePage object by parsing the annotated html"""
    parser = TemplatePageParser(token_dict)
    parser.feed(template_html)
    return parser.to_template()

def parse_extraction_page(token_dict, page_html):
    """Create an ExtractionPage object by parsing the html"""
    parser = ExtractionPageParser(token_dict)
    parser.feed(page_html)
    return parser.to_extraction_page()

class InstanceLearningParser(object):
    """Base parser for instance based learning algorithm
    
    This does not require correct HTML and the parsing method should not alter
    the original tag order. It is important that parsing results do not vary.
    """
    def __init__(self, token_dict):
        self.token_dict = token_dict
        self.token_list = []
    
    def _add_token(self, token, token_type, start, end):
        tid = self.token_dict.tokenid(token, token_type)
        self.token_list.append(tid)

    def feed(self, html_page):
        self.html_page = html_page
        self.previous_element_class = None
        for data in html_page.parsed_body:
            if isinstance(data, HtmlTag):
                self._add_token(data.tag, data.tag_type, data.start, data.end)
                self.handle_tag(data)
            else:
                self.handle_data(data)
            self.previous_element_class = data.__class__

    def handle_data(self, html_data_fragment):
        pass

    def handle_tag(self, html_tag):
        pass

_END_UNPAIREDTAG_TAGS = ["form", "div", "p", "table", "tr", "td"]

class TemplatePageParser(InstanceLearningParser):
    """Template parsing for instance based learning algorithm"""

    def __init__(self, token_dict):
        InstanceLearningParser.__init__(self, token_dict)
        self.annotations = []
        self.ignored_regions = []
        self.extra_required_attrs = []
        self.ignored_tag_stacks = defaultdict(list)
        # tag names that have not been completed
        self.labelled_tag_stacks = defaultdict(list)
        self.replacement_stacks = defaultdict(list)
        self.unpairedtag_stack = []
        self.variant_stack = []
        self.prev_data = None
        self.last_text_region = None
        self.next_tag_index = 0

    def handle_tag(self, html_tag):
        if self.last_text_region:
            self._process_text('')
        
        if html_tag.tag_type == HtmlTagType.OPEN_TAG:
            self._handle_open_tag(html_tag)
        elif html_tag.tag_type == HtmlTagType.CLOSE_TAG:
            self._handle_close_tag(html_tag)
        else:
            # the tag is not paired, it can contain only attribute annotations
            self._handle_unpaired_tag(html_tag)
    
    @staticmethod
    def _read_template_annotation(html_tag):
        template_attr = html_tag.attributes.get('data-scrapy-annotate')
        if template_attr is None:
            return None
        unescaped = template_attr.replace('&quot;', '"')
        return json.loads(unescaped)
    
    @staticmethod
    def _read_bool_template_attribute(html_tag, attribute):
        return html_tag.attributes.get("data-scrapy-" + attribute) == "true"
    
    def _close_unpaired_tag(self):
        self.unpairedtag_stack[0].end_index = self.next_tag_index
        self.unpairedtag_stack = []

    def _handle_unpaired_tag(self, html_tag):
        if self._read_bool_template_attribute(html_tag, "ignore") and html_tag.tag == "img":
            self.ignored_regions.append((self.next_tag_index, self.next_tag_index + 1))
        elif self._read_bool_template_attribute(html_tag, "ignore-beneath"):
            self.ignored_regions.append((self.next_tag_index, None))
        jannotation = self._read_template_annotation(html_tag)
        if jannotation:
            if self.unpairedtag_stack:
                self._close_unpaired_tag()
                
            annotation = AnnotationTag(self.next_tag_index, self.next_tag_index + 1)
            attribute_annotations = jannotation.pop('annotations', {}).items()
            for extract_attribute, tag_value in attribute_annotations:
                if extract_attribute == 'content':
                    annotation.surrounds_attribute = tag_value
                    self.unpairedtag_stack.append(annotation)
                else:
                    annotation.tag_attributes.append((extract_attribute, tag_value))
            self.annotations.append(annotation)
            if jannotation.pop('common_prefix', False):
                annotation.match_common_prefix = True

            self.extra_required_attrs.extend(jannotation.pop('required', []))
            annotation.metadata = jannotation

        self.next_tag_index += 1

    def _handle_open_tag(self, html_tag):
        if self._read_bool_template_attribute(html_tag, "ignore"):
            if html_tag.tag == "img":
                self.ignored_regions.append((self.next_tag_index, self.next_tag_index + 1))
            else:
                self.ignored_regions.append((self.next_tag_index, None))
                self.ignored_tag_stacks[html_tag.tag].append(html_tag)
                
        elif self.ignored_tag_stacks.get(html_tag.tag):
            self.ignored_tag_stacks[html_tag.tag].append(None)
        if self._read_bool_template_attribute(html_tag, "ignore-beneath"):
            self.ignored_regions.append((self.next_tag_index, None))
        
        replacement = html_tag.attributes.pop("data-scrapy-replacement", None)
        if replacement:
            self.token_list.pop()
            self._add_token(replacement, html_tag.tag_type, html_tag.start, html_tag.end)
            self.replacement_stacks[html_tag.tag].append(replacement)
        elif html_tag.tag in self.replacement_stacks:
            self.replacement_stacks[html_tag.tag].append(None)

        if self.unpairedtag_stack:
            if html_tag.tag in _END_UNPAIREDTAG_TAGS:
                self._close_unpaired_tag()
            else:
                self.unpairedtag_stack.append(html_tag.tag)
            
        # can't be a p inside another p. Also, an open p element closes
        # a previous open p element.
        if html_tag.tag == "p" and html_tag.tag in self.labelled_tag_stacks:
            annotation = self.labelled_tag_stacks.pop(html_tag.tag)[0]
            annotation.end_index = self.next_tag_index
            self.annotations.append(annotation)
                
        jannotation = self._read_template_annotation(html_tag)
        if not jannotation:
            if html_tag.tag in self.labelled_tag_stacks:
                # add this tag to the stack to match correct end tag
                self.labelled_tag_stacks[html_tag.tag].append(None)
            self.next_tag_index += 1
            return
        
        annotation = AnnotationTag(self.next_tag_index, None)
        if jannotation.pop('generated', False):
            self.token_list.pop()
            annotation.start_index -= 1
            if self.previous_element_class == HtmlTag:
                annotation.annotation_text = AnnotationText('')
            else:
                annotation.annotation_text = AnnotationText(self.prev_data)
            if self._read_bool_template_attribute(html_tag, "ignore") \
                    or self._read_bool_template_attribute(html_tag, "ignore-beneath"):
                ignored = self.ignored_regions.pop()
                self.ignored_regions.append((ignored[0]-1, ignored[1]))
                
        if jannotation.pop('common_prefix', False):
            annotation.match_common_prefix = True
                
        self.extra_required_attrs.extend(jannotation.pop('required', []))
        
        variant_id = jannotation.pop('variant', 0)
        if variant_id > 0:
            self.variant_stack.append(variant_id)
            annotation.surrounds_variant = variant_id
        attribute_annotations = jannotation.pop('annotations', {}).items()
        for extract_attribute, tag_value in attribute_annotations:
            if extract_attribute == 'content':
                annotation.surrounds_attribute = tag_value
            else:
                annotation.tag_attributes.append((extract_attribute, tag_value))
        
        annotation.metadata = jannotation

        if annotation.annotation_text is None:
            self.next_tag_index += 1
        if self.variant_stack:
            variant_id = self.variant_stack[-1]
            if variant_id == '0':
                variant_id = None
            annotation.variant_id = variant_id
        
        # look for a closing tag if the content is important
        if annotation.surrounds_attribute or annotation.surrounds_variant:
            self.labelled_tag_stacks[html_tag.tag].append(annotation)
        else:
            annotation.end_index = annotation.start_index + 1
            self.annotations.append(annotation)

    def _handle_close_tag(self, html_tag):
        
        if self.unpairedtag_stack:
            if html_tag.tag == self.unpairedtag_stack[-1]:
                self.unpairedtag_stack.pop()
            else:
                self._close_unpaired_tag()
        ignored_tags = self.ignored_tag_stacks.get(html_tag.tag)
        if ignored_tags is not None:
            tag = ignored_tags.pop()
            if isinstance(tag, HtmlTag):
                for i in range(-1, -len(self.ignored_regions) - 1, -1):
                    if self.ignored_regions[i][1] is None:
                        self.ignored_regions[i] = (self.ignored_regions[i][0], self.next_tag_index)
                        break
            if len(ignored_tags) == 0:
                del self.ignored_tag_stacks[html_tag.tag]

        if html_tag.tag in self.replacement_stacks:
            replacement = self.replacement_stacks[html_tag.tag].pop()
            if replacement:
                self.token_list.pop()
                self._add_token(replacement, html_tag.tag_type, html_tag.start, html_tag.end)
            if len(self.replacement_stacks[html_tag.tag]) == 0:
                del self.replacement_stacks[html_tag.tag]

        labelled_tags = self.labelled_tag_stacks.get(html_tag.tag)
        if labelled_tags is None:
            self.next_tag_index += 1
            return
        annotation = labelled_tags.pop()
        if annotation is None:
            self.next_tag_index += 1
        else:
            annotation.end_index = self.next_tag_index
            self.annotations.append(annotation)
            if annotation.annotation_text is not None:
                self.token_list.pop()
                self.last_text_region = annotation
            else:
                self.next_tag_index += 1
            if len(labelled_tags) == 0:
                del self.labelled_tag_stacks[html_tag.tag]
            if annotation.surrounds_variant and self.variant_stack:
                prev = self.variant_stack.pop()
                if prev != annotation.surrounds_variant:
                    raise ValueError("unbalanced variant annotation tags")
                    
    def handle_data(self, html_data_fragment):
        fragment_text = self.html_page.fragment_data(html_data_fragment)
        self._process_text(fragment_text)

    def _process_text(self, text):
        if self.last_text_region is not None:
            self.last_text_region.annotation_text.follow_text = text
            self.last_text_region = None
        self.prev_data = text

    def to_template(self):
        """create a TemplatePage from the data fed to this parser"""
        return TemplatePage(self.token_dict, self.token_list, self.annotations,
                self.html_page.page_id, self.ignored_regions, self.extra_required_attrs)

class ExtractionPageParser(InstanceLearningParser):
    """Parse an HTML page for extraction using the instance based learning
    algorithm

    This needs to extract the tokens in a similar way to LabelledPageParser,
    it needs to also maintain a mapping from token index to the original content
    so that once regions are identified, the original content can be extracted.
    """
    def __init__(self, token_dict):
        InstanceLearningParser.__init__(self, token_dict)
        self.page_data = []
        self.token_start_index = []
        self.token_follow_index = []
        self.tag_attrs = {}

    def _add_token(self, token, token_type, start, end):
        InstanceLearningParser._add_token(self, token, token_type, start, end)
        self.token_start_index.append(start)
        self.token_follow_index.append(end)

    def handle_tag(self, html_tag):
        if html_tag.attributes:
            self.tag_attrs[len(self.token_list) - 1] = html_tag.attributes
    
    def to_extraction_page(self):
        return ExtractionPage(self.html_page.body, self.token_dict, array(self.token_list), 
                self.token_start_index, self.token_follow_index, self.tag_attrs)



"""
IBL module

This contains an extraction algorithm based on the paper Extracting Web Data
Using Instance-Based Learning by Yanhong Zhai and Bing Liu.

It defines the InstanceBasedLearningExtractor class, which implements this
extraction algorithm.

Main departures from the original algorithm:
    * there is no limit in prefix or suffix size
    * we have "attribute adaptors" that allow generic post processing and may
      affect the extraction process. For example, a price field may require a 
      numeric value to be present.
    * tags can be inserted to extract regions not wrapped by html tags. These
      regions are then identified using the longest unique character prefix and
      suffix.
"""
from operator import itemgetter
from .regionextract import build_extraction_tree
from .pageparsing import parse_template, parse_extraction_page
from .pageobjects import TokenDict, AnnotationText
from .similarity import common_prefix

class InstanceBasedLearningExtractor(object):
    """Implementation of the instance based learning algorithm to 
    extract data from web pages.
    """

    def __init__(self, templates, type_descriptor=None, trace=False):
        """Initialise this extractor

        templates should contain a sequence of strings, each containing 
        annotated html that will be used as templates for extraction.

        Tags surrounding areas to be extracted must contain a 
        'data-scrapy-annotate' attribute and the value must be the name
        of the attribute. If the tag was inserted and was not present in the 
        original page, the data-scrapy-generated attribute must be present.
        
        type_descriptor may contain a type descriptor describing the item
        to be extracted.
        
        if trace is true, the returned extracted data will have a 'trace'
        property that contains a trace of the extraction execution.
        """
        self.token_dict = TokenDict()
        parsed_plus_templates = [(parse_template(self.token_dict, t), t) for t in templates]
        parsed_plus_epages = [(p, parse_extraction_page(self.token_dict, t)) for p, t \
               in parsed_plus_templates if _annotation_count(p)]
        parsed_templates = map(itemgetter(0), parsed_plus_epages)
        
        # calculate common text prefixes for annotations of same field across all templates
        extracted_text = {}
        extraction_pages = map(itemgetter(1), parsed_plus_epages)
        for i, parsed in enumerate(parsed_templates):
            for annot in parsed.annotations:
                if annot.match_common_prefix:
                    field = annot.surrounds_attribute
                    if field is not None:
                        descriptor = type_descriptor.attribute_map.get(field) if type_descriptor else None
                        allow_markup = descriptor.allow_markup if descriptor else False
                        start, end = annot.start_index, annot.end_index
                        if allow_markup:
                            text = extraction_pages[i].html_between_tokens(start, end)
                        else:
                            text = extraction_pages[i].text_between_tokens(start, end)
                        extracted_text.setdefault(field, []).append(text)
        common_prefixes = {}
        for field, data in extracted_text.iteritems():
            if len(data) > 1:
                cprefix = common_prefix(*data)
                if cprefix:
                    common_prefixes[field] = "".join(cprefix).strip()
        # now apply common prefixes to annotations
        for i, parsed in enumerate(parsed_templates):
            for annot in parsed.annotations:
                for field, prefix in common_prefixes.iteritems():
                    if annot.surrounds_attribute == field and not annot.annotation_text:
                        annot.annotation_text = AnnotationText(prefix)
        
        # templates with more attributes are considered first
        sorted_templates = sorted(parsed_templates, key=_annotation_count, reverse=True)
        self.extraction_trees = [build_extraction_tree(t, type_descriptor, 
            trace) for t in sorted_templates]
        self.validated = type_descriptor.validated if type_descriptor else \
                self._filter_not_none

    def extract(self, html, pref_template_id=None, useone=False):
        """extract data from an html page
        
        If pref_template_url is specified, the template with that url will be 
        used first.
        if useone is True and no data was extracted, no additional template will
        be tried. If False and no data was extracted, try with rest of item templates
        """
        extraction_page = parse_extraction_page(self.token_dict, html)
        if pref_template_id is not None:
            if useone:
                extraction_trees = [x for x in self.extraction_trees if x.template.id == pref_template_id]
            else:
                extraction_trees = sorted(self.extraction_trees, 
                    key=lambda x: x.template.id != pref_template_id)
        else:
            extraction_trees = self.extraction_trees

        for extraction_tree in extraction_trees:
            extracted = extraction_tree.extract(extraction_page)
            correctly_extracted = self.validated(extracted)
            extra_required = extraction_tree.template.extra_required_attrs
            correctly_extracted = [c for c in correctly_extracted if \
                extra_required.intersection(c.keys()) == extra_required ]
            if len(correctly_extracted) > 0:
                return correctly_extracted, extraction_tree.template
        return None, None

    def __str__(self):
        return "InstanceBasedLearningExtractor[\n%s\n]" % \
                (',\n'.join(map(str, self.extraction_trees)))
    
    @staticmethod
    def _filter_not_none(items):
        return [d for d in items if d is not None]

def _annotation_count(template):
    return len(template.annotations)



import os
import unittest
import urlparse

from scrapy.xlib.BeautifulSoup import BeautifulSoup
from scrapy.http import Response, TextResponse, HtmlResponse
from scrapy.utils.response import body_or_str, get_base_url, get_meta_refresh, \
    response_httprepr, get_cached_beautifulsoup, open_in_browser

class ResponseUtilsTest(unittest.TestCase):
    dummy_response = TextResponse(url='http://example.org/', body='dummy_response')

    def test_body_or_str_input(self):
        self.assertTrue(isinstance(body_or_str(self.dummy_response), basestring))
        self.assertTrue(isinstance(body_or_str('text'), basestring))
        self.assertRaises(Exception, body_or_str, 2)

    def test_body_or_str_extraction(self):
        self.assertEqual(body_or_str(self.dummy_response), 'dummy_response')
        self.assertEqual(body_or_str('text'), 'text')

    def test_body_or_str_encoding(self):
        self.assertTrue(isinstance(body_or_str(self.dummy_response, unicode=False), str))
        self.assertTrue(isinstance(body_or_str(self.dummy_response, unicode=True), unicode))

        self.assertTrue(isinstance(body_or_str('text', unicode=False), str))
        self.assertTrue(isinstance(body_or_str('text', unicode=True), unicode))

        self.assertTrue(isinstance(body_or_str(u'text', unicode=False), str))
        self.assertTrue(isinstance(body_or_str(u'text', unicode=True), unicode))

    def test_get_base_url(self):
        response = HtmlResponse(url='https://example.org', body="""\
            <html>\
            <head><title>Dummy</title><base href='http://example.org/something' /></head>\
            <body>blahablsdfsal&amp;</body>\
            </html>""")
        self.assertEqual(get_base_url(response), 'http://example.org/something')

        # relative url with absolute path
        response = HtmlResponse(url='https://example.org', body="""\
            <html>\
            <head><title>Dummy</title><base href='/absolutepath' /></head>\
            <body>blahablsdfsal&amp;</body>\
            </html>""")
        self.assertEqual(get_base_url(response), 'https://example.org/absolutepath')

        # no scheme url
        response = HtmlResponse(url='https://example.org', body="""\
            <html>\
            <head><title>Dummy</title><base href='//noscheme.com/path' /></head>\
            <body>blahablsdfsal&amp;</body>\
            </html>""")
        self.assertEqual(get_base_url(response), 'https://noscheme.com/path')

    def test_get_meta_refresh(self):
        body = """
            <html>
            <head><title>Dummy</title><meta http-equiv="refresh" content="5;url=http://example.org/newpage" /></head>
            <body>blahablsdfsal&amp;</body>
            </html>"""
        response = TextResponse(url='http://example.org', body=body)
        self.assertEqual(get_meta_refresh(response), (5, 'http://example.org/newpage'))

        # refresh without url should return (None, None)
        body = """<meta http-equiv="refresh" content="5" />"""
        response = TextResponse(url='http://example.org', body=body)
        self.assertEqual(get_meta_refresh(response), (None, None))

        body = """<meta http-equiv="refresh" content="5;
            url=http://example.org/newpage" /></head>"""
        response = TextResponse(url='http://example.org', body=body)
        self.assertEqual(get_meta_refresh(response), (5, 'http://example.org/newpage'))

        # meta refresh in multiple lines
        body = """<html><head>
               <META
               HTTP-EQUIV="Refresh"
               CONTENT="1; URL=http://example.org/newpage">"""
        response = TextResponse(url='http://example.org', body=body)
        self.assertEqual(get_meta_refresh(response), (1, 'http://example.org/newpage'))

        # entities in the redirect url
        body = """<meta http-equiv="refresh" content="3; url=&#39;http://www.example.com/other&#39;">"""
        response = TextResponse(url='http://example.com', body=body)
        self.assertEqual(get_meta_refresh(response), (3, 'http://www.example.com/other'))

        # relative redirects
        body = """<meta http-equiv="refresh" content="3; url=other.html">"""
        response = TextResponse(url='http://example.com/page/this.html', body=body)
        self.assertEqual(get_meta_refresh(response), (3, 'http://example.com/page/other.html'))

        # non-standard encodings (utf-16)
        body = """<meta http-equiv="refresh" content="3; url=http://example.com/redirect">"""
        body = body.decode('ascii').encode('utf-16')
        response = TextResponse(url='http://example.com', body=body, encoding='utf-16')
        self.assertEqual(get_meta_refresh(response), (3, 'http://example.com/redirect'))

        # non-ascii chars in the url (utf8)
        body = """<meta http-equiv="refresh" content="3; url=http://example.com/to\xc2\xa3">"""
        response = TextResponse(url='http://example.com', body=body, encoding='utf-8')
        self.assertEqual(get_meta_refresh(response), (3, 'http://example.com/to%C2%A3'))

        # non-ascii chars in the url (latin1)
        body = """<meta http-equiv="refresh" content="3; url=http://example.com/to\xa3">"""
        response = TextResponse(url='http://example.com', body=body, encoding='latin1')
        self.assertEqual(get_meta_refresh(response), (3, 'http://example.com/to%C2%A3'))

        # responses without refresh tag should return None None
        response = TextResponse(url='http://example.org')
        self.assertEqual(get_meta_refresh(response), (None, None))
        response = TextResponse(url='http://example.org')
        self.assertEqual(get_meta_refresh(response), (None, None))

        # html commented meta refresh header must not directed
        body = """<!--<meta http-equiv="refresh" content="3; url=http://example.com/">-->"""
        response = TextResponse(url='http://example.com', body=body)
        self.assertEqual(get_meta_refresh(response), (None, None))

        # html comments must not interfere with uncommented meta refresh header
        body = """<!-- commented --><meta http-equiv="refresh" content="3; url=http://example.com/">-->"""
        response = TextResponse(url='http://example.com', body=body)
        self.assertEqual(get_meta_refresh(response), (3, 'http://example.com/'))

        # float refresh intervals
        body = """<meta http-equiv="refresh" content=".1;URL=index.html" />"""
        response = TextResponse(url='http://example.com', body=body)
        self.assertEqual(get_meta_refresh(response), (0.1, 'http://example.com/index.html'))

        body = """<meta http-equiv="refresh" content="3.1;URL=index.html" />"""
        response = TextResponse(url='http://example.com', body=body)
        self.assertEqual(get_meta_refresh(response), (3.1, 'http://example.com/index.html'))

    def test_response_httprepr(self):
        r1 = Response("http://www.example.com")
        self.assertEqual(response_httprepr(r1), 'HTTP/1.1 200 OK\r\n\r\n')

        r1 = Response("http://www.example.com", status=404, headers={"Content-type": "text/html"}, body="Some body")
        self.assertEqual(response_httprepr(r1), 'HTTP/1.1 404 Not Found\r\nContent-Type: text/html\r\n\r\nSome body')

        r1 = Response("http://www.example.com", status=6666, headers={"Content-type": "text/html"}, body="Some body")
        self.assertEqual(response_httprepr(r1), 'HTTP/1.1 6666 \r\nContent-Type: text/html\r\n\r\nSome body')

    def test_get_cached_beautifulsoup(self):
        r1 = Response('http://www.example.com', body='')

        soup1 = get_cached_beautifulsoup(r1)
        soup2 = get_cached_beautifulsoup(r1)

        assert isinstance(soup1, BeautifulSoup)
        assert isinstance(soup2, BeautifulSoup)
        # make sure it's cached
        assert soup1 is soup2

        # when body is None, an empty soup should be returned
        r1 = Response('http://www.example.com')
        assert r1.body == ""
        assert isinstance(get_cached_beautifulsoup(r1), BeautifulSoup)

        r1 = Response('http://www.example.com', body='')
        soup1 = get_cached_beautifulsoup(r1)
        r2 = r1.copy()
        soup2 = get_cached_beautifulsoup(r1)
        soup3 = get_cached_beautifulsoup(r2)

        assert soup1 is soup2
        assert soup1 is not soup3

    def test_open_in_browser(self):
        url = "http:///www.example.com/some/page.html"
        body = "<html> <head> <title>test page</title> </head> <body>test body</body> </html>"
        def browser_open(burl):
            path = urlparse.urlparse(burl).path
            if not os.path.exists(path):
                path = burl.replace('file://', '')
            bbody = open(path).read()
            assert '<base href="%s">' % url in bbody, "<base> tag not added"
            return True
        response = HtmlResponse(url, body=body)
        assert open_in_browser(response, _openfunc=browser_open), \
            "Browser not called"
        self.assertRaises(TypeError, open_in_browser, Response(url, body=body), \
            debug=True)

if __name__ == "__main__":
    unittest.main()



import scrapy
from scrapy.command import ScrapyCommand

class Command(ScrapyCommand):

    def short_desc(self):
        return "Print Scrapy version"

    def run(self, args, opts):
        print "Scrapy %s" % scrapy.__version__



import os
from ConfigParser import NoSectionError

from scrapy.spiderqueue import SqliteSpiderQueue

def get_spider_queues(config):
    """Return a dict of Spider Quees keyed by project name"""
    dbsdir = config.get('dbs_dir', 'dbs')
    if not os.path.exists(dbsdir):
        os.makedirs(dbsdir)
    d = {}
    for project in get_project_list(config):
        dbpath = os.path.join(dbsdir, '%s.db' % project)
        d[project] = SqliteSpiderQueue(dbpath)
    return d

def get_project_list(config):
    """Get list of projects by inspecting the eggs dir and the ones defined in
    the scrapyd.conf [settings] section
    """
    eggs_dir = config.get('eggs_dir', 'eggs')
    if os.path.exists(eggs_dir):
        projects = os.listdir(eggs_dir)
    else:
        projects = []
    try:
        projects += [x[0] for x in config.cp.items('settings')]
    except NoSectionError:
        pass
    return projects



import re
import unittest
import weakref

import libxml2

from scrapy.http import TextResponse, HtmlResponse, XmlResponse
from scrapy.selector import XmlXPathSelector, HtmlXPathSelector, \
    XPathSelector
from scrapy.selector.document import Libxml2Document
from scrapy.utils.test import libxml2debug

class XPathSelectorTestCase(unittest.TestCase):

    @libxml2debug
    def test_selector_simple(self):
        """Simple selector tests"""
        body = "<p><input name='a'value='1'/><input name='b'value='2'/></p>"
        response = TextResponse(url="http://example.com", body=body)
        xpath = HtmlXPathSelector(response)

        xl = xpath.select('//input')
        self.assertEqual(2, len(xl))
        for x in xl:
            assert isinstance(x, HtmlXPathSelector)

        self.assertEqual(xpath.select('//input').extract(),
                         [x.extract() for x in xpath.select('//input')])

        self.assertEqual([x.extract() for x in xpath.select("//input[@name='a']/@name")],
                         [u'a'])
        self.assertEqual([x.extract() for x in xpath.select("number(concat(//input[@name='a']/@value, //input[@name='b']/@value))")],
                         [u'12.0'])

        self.assertEqual(xpath.select("concat('xpath', 'rules')").extract(),
                         [u'xpathrules'])
        self.assertEqual([x.extract() for x in xpath.select("concat(//input[@name='a']/@value, //input[@name='b']/@value)")],
                         [u'12'])

    @libxml2debug
    def test_selector_same_type(self):
        """Test XPathSelector returning the same type in x() method"""
        text = '<p>test<p>'
        assert isinstance(XmlXPathSelector(text=text).select("//p")[0],
                          XmlXPathSelector)
        assert isinstance(HtmlXPathSelector(text=text).select("//p")[0], 
                          HtmlXPathSelector)

    @libxml2debug
    def test_selector_xml_html(self):
        """Test that XML and HTML XPathSelector's behave differently"""

        # some text which is parsed differently by XML and HTML flavors
        text = '<div><img src="a.jpg"><p>Hello</div>'

        self.assertEqual(XmlXPathSelector(text=text).select("//div").extract(),
                         [u'<div><img src="a.jpg"><p>Hello</p></img></div>'])

        self.assertEqual(HtmlXPathSelector(text=text).select("//div").extract(),
                         [u'<div><img src="a.jpg"><p>Hello</p></div>'])

    @libxml2debug
    def test_selector_nested(self):
        """Nested selector tests"""
        body = """<body>
                    <div class='one'>
                      <ul>
                        <li>one</li><li>two</li>
                      </ul>
                    </div>
                    <div class='two'>
                      <ul>
                        <li>four</li><li>five</li><li>six</li>
                      </ul>
                    </div>
                  </body>"""

        response = HtmlResponse(url="http://example.com", body=body)
        x = HtmlXPathSelector(response)

        divtwo = x.select('//div[@class="two"]')
        self.assertEqual(divtwo.select("//li").extract(),
                         ["<li>one</li>", "<li>two</li>", "<li>four</li>", "<li>five</li>", "<li>six</li>"])
        self.assertEqual(divtwo.select("./ul/li").extract(),
                         ["<li>four</li>", "<li>five</li>", "<li>six</li>"])
        self.assertEqual(divtwo.select(".//li").extract(),
                         ["<li>four</li>", "<li>five</li>", "<li>six</li>"])
        self.assertEqual(divtwo.select("./li").extract(),
                         [])

    @libxml2debug
    def test_selector_re(self):
        body = """<div>Name: Mary
                    <ul>
                      <li>Name: John</li>
                      <li>Age: 10</li>
                      <li>Name: Paul</li>
                      <li>Age: 20</li>
                    </ul>
                    Age: 20
                  </div>

               """
        response = HtmlResponse(url="http://example.com", body=body)
        x = HtmlXPathSelector(response)

        name_re = re.compile("Name: (\w+)")
        self.assertEqual(x.select("//ul/li").re(name_re),
                         ["John", "Paul"])
        self.assertEqual(x.select("//ul/li").re("Age: (\d+)"),
                         ["10", "20"])

    @libxml2debug
    def test_selector_over_text(self):
        hxs = HtmlXPathSelector(text='<root>lala</root>')
        self.assertEqual(hxs.extract(),
                         u'<html><body><root>lala</root></body></html>')

        xxs = XmlXPathSelector(text='<root>lala</root>')
        self.assertEqual(xxs.extract(),
                         u'<root>lala</root>')

        xxs = XmlXPathSelector(text='<root>lala</root>')
        self.assertEqual(xxs.select('.').extract(),
                         [u'<root>lala</root>'])


    @libxml2debug
    def test_selector_namespaces_simple(self):
        body = """
        <test xmlns:somens="http://scrapy.org">
           <somens:a id="foo"/>
           <a id="bar">found</a>
        </test>
        """

        response = XmlResponse(url="http://example.com", body=body)
        x = XmlXPathSelector(response)
        
        x.register_namespace("somens", "http://scrapy.org")
        self.assertEqual(x.select("//somens:a").extract(), 
                         ['<somens:a id="foo"/>'])


    @libxml2debug
    def test_selector_namespaces_multiple(self):
        body = """<?xml version="1.0" encoding="UTF-8"?>
<BrowseNode xmlns="http://webservices.amazon.com/AWSECommerceService/2005-10-05"
            xmlns:b="http://somens.com"
            xmlns:p="http://www.scrapy.org/product" >
    <b:Operation>hello</b:Operation>
    <TestTag b:att="value"><Other>value</Other></TestTag>
    <p:SecondTestTag><material/><price>90</price><p:name>Dried Rose</p:name></p:SecondTestTag>
</BrowseNode>
        """
        response = XmlResponse(url="http://example.com", body=body)
        x = XmlXPathSelector(response)

        x.register_namespace("xmlns", "http://webservices.amazon.com/AWSECommerceService/2005-10-05")
        x.register_namespace("p", "http://www.scrapy.org/product")
        x.register_namespace("b", "http://somens.com")
        self.assertEqual(len(x.select("//xmlns:TestTag")), 1)
        self.assertEqual(x.select("//b:Operation/text()").extract()[0], 'hello')
        self.assertEqual(x.select("//xmlns:TestTag/@b:att").extract()[0], 'value')
        self.assertEqual(x.select("//p:SecondTestTag/xmlns:price/text()").extract()[0], '90')
        self.assertEqual(x.select("//p:SecondTestTag").select("./xmlns:price/text()")[0].extract(), '90')
        self.assertEqual(x.select("//p:SecondTestTag/xmlns:material").extract()[0], '<material/>')

    @libxml2debug
    def test_selector_invalid_xpath(self):
        response = XmlResponse(url="http://example.com", body="<html></html>")
        x = HtmlXPathSelector(response)
        xpath = "//test[@foo='bar]"
        try:
            x.select(xpath)
        except ValueError, e:
            assert xpath in str(e), "Exception message does not contain invalid xpath"
        except Exception:
            raise AssertionError("A invalid XPath does not raise ValueError")
        else:
            raise AssertionError("A invalid XPath does not raise an exception")

    @libxml2debug
    def test_http_header_encoding_precedence(self):
        # u'\xa3'     = pound symbol in unicode
        # u'\xc2\xa3' = pound symbol in utf-8
        # u'\xa3'     = pound symbol in latin-1 (iso-8859-1)

        meta = u'<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">'
        head = u'<head>' + meta + u'</head>'
        body_content = u'<span id="blank">\xa3</span>'
        body = u'<body>' + body_content + u'</body>'
        html = u'<html>' + head + body + u'</html>'
        encoding = 'utf-8'
        html_utf8 = html.encode(encoding)

        headers = {'Content-Type': ['text/html; charset=utf-8']}
        response = HtmlResponse(url="http://example.com", headers=headers, body=html_utf8)
        x = HtmlXPathSelector(response)
        self.assertEquals(x.select("//span[@id='blank']/text()").extract(),
                          [u'\xa3'])

    @libxml2debug
    def test_null_bytes(self):
        hxs = HtmlXPathSelector(text='<root>la\x00la</root>')
        self.assertEqual(hxs.extract(),
                         u'<html><body><root>lala</root></body></html>')

        xxs = XmlXPathSelector(text='<root>la\x00la</root>')
        self.assertEqual(xxs.extract(),
                         u'<root>lala</root>')

    @libxml2debug
    def test_unquote(self):
        xmldoc = '\n'.join((
            '<root>',
            '  lala',
            '  <node>',
            '    blabla&amp;more<!--comment-->a<b>test</b>oh',
            '    <![CDATA[lalalal&ppppp<b>PPPP</b>ppp&amp;la]]>',
            '  </node>',
            '  pff',
            '</root>'))
        xxs = XmlXPathSelector(text=xmldoc)

        self.assertEqual(xxs.extract_unquoted(), u'')

        self.assertEqual(xxs.select('/root').extract_unquoted(), [u''])
        self.assertEqual(xxs.select('/root/text()').extract_unquoted(), [
            u'\n  lala\n  ',
            u'\n  pff\n'])

        self.assertEqual(xxs.select('//*').extract_unquoted(), [u'', u'', u''])
        self.assertEqual(xxs.select('//text()').extract_unquoted(), [
            u'\n  lala\n  ',
            u'\n    blabla&more',
            u'a',
            u'test',
            u'oh\n    ',
            u'lalalal&ppppp<b>PPPP</b>ppp&amp;la',
            u'\n  ',
            u'\n  pff\n'])

    @libxml2debug
    def test_empty_bodies(self):
        r1 = TextResponse('http://www.example.com', body='')
        hxs = HtmlXPathSelector(r1) # shouldn't raise error
        xxs = XmlXPathSelector(r1) # shouldn't raise error

    @libxml2debug
    def test_weakref_slots(self):
        """Check that classes are using slots and are weak-referenceable"""
        for cls in [XPathSelector, HtmlXPathSelector, XmlXPathSelector]:
            x = cls()
            weakref.ref(x)
            assert not hasattr(x, '__dict__'), "%s does not use __slots__" % \
                x.__class__.__name__

class Libxml2DocumentTest(unittest.TestCase):

    @libxml2debug
    def test_response_libxml2_caching(self):
        r1 = HtmlResponse('http://www.example.com', body='<html><head></head><body></body></html>')
        r2 = r1.copy()

        doc1 = Libxml2Document(r1)
        doc2 = Libxml2Document(r1)
        doc3 = Libxml2Document(r2)

        # make sure it's cached
        assert doc1 is doc2
        assert doc1.xmlDoc is doc2.xmlDoc
        assert doc1 is not doc3
        assert doc1.xmlDoc is not doc3.xmlDoc

        # don't leave libxml2 documents in memory to avoid wrong libxml2 leaks reports
        del doc1, doc2, doc3

    @libxml2debug
    def test_null_char(self):
        # make sure bodies with null char ('\x00') don't raise a TypeError exception
        self.body_content = 'test problematic \x00 body'
        response = TextResponse('http://example.com/catalog/product/blabla-123',
                            headers={'Content-Type': 'text/plain; charset=utf-8'}, body=self.body_content)
        Libxml2Document(response)

class Libxml2Test(unittest.TestCase):

    @libxml2debug
    def test_libxml2_bug_2_6_27(self):
        # this test will fail in version 2.6.27 but passes on 2.6.29+
        html = "<td>1<b>2</b>3</td>"
        node = libxml2.htmlParseDoc(html, 'utf-8')
        result = [str(r) for r in node.xpathEval('//text()')]
        self.assertEquals(result, ['1', '2', '3'])
        node.freeDoc()

if __name__ == "__main__":
    unittest.main()



from collections import defaultdict
from scrapy.xlib.pydispatch import dispatcher

from scrapy import signals
from scrapy.http import Response
from scrapy.http.cookies import CookieJar
from scrapy.conf import settings
from scrapy import log


class CookiesMiddleware(object):
    """This middleware enables working with sites that need cookies"""
    debug = settings.getbool('COOKIES_DEBUG')

    def __init__(self):
        self.jars = defaultdict(CookieJar)
        dispatcher.connect(self.spider_closed, signals.spider_closed)

    def process_request(self, request, spider):
        if 'dont_merge_cookies' in request.meta:
            return

        jar = self.jars[spider]
        cookies = self._get_request_cookies(jar, request)
        for cookie in cookies:
            jar.set_cookie_if_ok(cookie, request)

        # set Cookie header
        request.headers.pop('Cookie', None)
        jar.add_cookie_header(request)
        self._debug_cookie(request)

    def process_response(self, request, response, spider):
        if 'dont_merge_cookies' in request.meta:
            return response

        # extract cookies from Set-Cookie and drop invalid/expired cookies
        jar = self.jars[spider]
        jar.extract_cookies(response, request)
        self._debug_set_cookie(response)

        return response

    def spider_closed(self, spider):
        self.jars.pop(spider, None)

    def _debug_cookie(self, request):
        """log Cookie header for request"""
        if self.debug:
            c = request.headers.get('Cookie')
            c = c and [p.split('=')[0] for p in c.split(';')]
            log.msg('Cookie: %s for %s' % (c, request.url), level=log.DEBUG)

    def _debug_set_cookie(self, response):
        """log Set-Cookies headers but exclude cookie values"""
        if self.debug:
            cl = response.headers.getlist('Set-Cookie')
            res = []
            for c in cl:
                kv, tail = c.split(';', 1)
                k = kv.split('=', 1)[0]
                res.append('%s %s' % (k, tail))
            log.msg('Set-Cookie: %s from %s' % (res, response.url))


    def _get_request_cookies(self, jar, request):
        headers = {'Set-Cookie': ['%s=%s;' % (k, v) for k, v in request.cookies.iteritems()]}
        response = Response(request.url, headers=headers)
        cookies = jar.make_cookies(response, request)
        return cookies





from twisted.internet import defer

from scrapy.http import Request
from scrapy.utils.misc import arg_to_iter
from scrapy.utils.spider import create_spider_for_request
from scrapy.utils.python import stringify_dict
from scrapy import log


class ExecutionQueue(object):

    def __init__(self, spiders, queue, poll_interval=5, keep_alive=False):
        self.spider_requests = []
        self.poll_interval = poll_interval
        self._spiders = spiders
        self._queue = queue
        self._keepalive = keep_alive

    @defer.inlineCallbacks
    def _append_next(self):
        """Called when there are no more items left in self.spider_requests.
        This method is meant to be overriden in subclasses to add new (spider,
        requests) tuples to self.spider_requests. It can return a Deferred.
        """
        msg = yield self._queue.pop()
        if msg:
            name = msg.pop('name')
            msg = stringify_dict(msg) # see #250
            self.append_spider_name(name, **msg)

    def get_next(self):
        """Return a tuple (spider, requests) containing a list of Requests and
        the Spider which will be used to crawl those Requests. If there aren't
        any more spiders to crawl it must return (None, []).

        This method can return a deferred.
        """
        if self.spider_requests:
            return self._get_next_now()
        d = defer.maybeDeferred(self._append_next)
        d.addCallback(lambda _: self._get_next_now())
        return d

    def _get_next_now(self):
        try:
            return self.spider_requests.pop(0)
        except IndexError:
            return (None, [])

    def is_finished(self):
        """Return True if the queue is empty and there won't be any more
        spiders to crawl (this is for one-shot runs). If it returns ``False``
        Scrapy will keep polling this queue for new requests to scrape
        """
        return not self._keepalive and not bool(self.spider_requests)

    def append_spider(self, spider):
        """Append a Spider to crawl"""
        requests = spider.start_requests()
        self.spider_requests.append((spider, requests))

    def append_request(self, request, spider=None, **kwargs):
        if spider is None:
            spider = create_spider_for_request(self._spiders, request, **kwargs)
        if spider:
            self.spider_requests.append((spider, [request]))

    def append_url(self, url=None, spider=None, **kwargs):
        """Append a URL to crawl with the given spider. If the spider is not
        given, a spider will be looked up based on the URL
        """
        if url is None:
            raise ValueError("A url is required")
        if spider is None:
            spider = create_spider_for_request(self._spiders, Request(url), \
                **kwargs)
        if spider:
            requests = arg_to_iter(spider.make_requests_from_url(url))
            self.spider_requests.append((spider, requests))

    def append_spider_name(self, name=None, **spider_kwargs):
        """Append a spider to crawl given its name and optional arguments,
        which are used to instantiate it. The SpiderManager is used to lookup
        the spider
        """
        if name is None:
            raise ValueError("A spider name is required")
        try:
            spider = self._spiders.create(name, **spider_kwargs)
        except KeyError:
            log.msg('Unable to find spider: %s' % name, log.ERROR)
        except:
            log.err(None, 'Error creating spider %r with arguments: %r' % \
                (name, spider_kwargs))
        else:
            self.append_spider(spider)



"""
Scrapy settings manager

See documentation in docs/topics/settings.rst
"""

import os
import cPickle as pickle

from scrapy.conf import default_settings
from scrapy.utils.conf import init_env

import_ = lambda x: __import__(x, {}, {}, [''])


class Settings(object):

    def __init__(self, values=None):
        self.values = values.copy() if values else {}
        self.global_defaults = default_settings

    def __getitem__(self, opt_name):
        if opt_name in self.values:
            return self.values[opt_name]
        return getattr(self.global_defaults, opt_name, None)

    def get(self, name, default=None):
        return self[name] if self[name] is not None else default

    def getbool(self, name, default=False):
        """
        True is: 1, '1', True
        False is: 0, '0', False, None
        """
        return bool(int(self.get(name, default)))

    def getint(self, name, default=0):
        return int(self.get(name, default))

    def getfloat(self, name, default=0.0):
        return float(self.get(name, default))

    def getlist(self, name, default=None):
        value = self.get(name)
        if value is None:
            return default or []
        elif hasattr(value, '__iter__'):
            return value
        else:
            return str(value).split(',')


class EnvironmentSettings(Settings):

    ENVVAR = 'SCRAPY_SETTINGS_MODULE'

    def __init__(self):
        super(EnvironmentSettings, self).__init__()
        self.defaults = {}
        self.disabled = os.environ.get('SCRAPY_SETTINGS_DISABLED', False)
        if self.ENVVAR not in os.environ:
            project = os.environ.get('SCRAPY_PROJECT', 'default')
            init_env(project)
        settings_module_path = os.environ.get(self.ENVVAR, 'scrapy_settings')
        self.set_settings_module(settings_module_path)

        # XXX: find a better solution for this hack
        pickled_settings = os.environ.get("SCRAPY_PICKLED_SETTINGS_TO_OVERRIDE")
        self.overrides = pickle.loads(pickled_settings) if pickled_settings else {}

    def set_settings_module(self, settings_module_path):
        self.settings_module_path = settings_module_path
        try:
            self.settings_module = import_(settings_module_path)
        except ImportError:
            self.settings_module = None

    def __getitem__(self, opt_name):
        if not self.disabled:
            if opt_name in self.overrides:
                return self.overrides[opt_name]
            if 'SCRAPY_' + opt_name in os.environ:
                return os.environ['SCRAPY_' + opt_name]
            if hasattr(self.settings_module, opt_name):
                return getattr(self.settings_module, opt_name)
            if opt_name in self.defaults:
                return self.defaults[opt_name]
        return super(EnvironmentSettings, self).__getitem__(opt_name)

    def __str__(self):
        return "<Settings %r>" % self.settings_module_path


settings = EnvironmentSettings()



from twisted.trial import unittest

from scrapy.http import HtmlResponse
from scrapy.spider import BaseSpider
from scrapy.contrib_exp.crawlspider.matchers import BaseMatcher
from scrapy.contrib_exp.crawlspider.matchers import UrlMatcher
from scrapy.contrib_exp.crawlspider.matchers import UrlRegexMatcher

from scrapy.contrib_exp.crawlspider.rules import CompiledRule
from scrapy.contrib_exp.crawlspider.rules import Rule
from scrapy.contrib_exp.crawlspider.rules import RulesManager

from functools import partial

class RuleInitializationTest(unittest.TestCase):

    def test_fail_if_rule_null(self):
        # fail on empty rule
        self.failUnlessRaises(ValueError, Rule)
        self.failUnlessRaises(ValueError, Rule,
                                **dict(callback=None, follow=None))
        self.failUnlessRaises(ValueError, Rule,
                                **dict(callback=None, follow=False))

    def test_minimal_arguments_to_instantiation(self):
        # not fail if callback set
        self.failUnless(Rule(callback=lambda: True))
        # not fail if follow set
        self.failUnless(Rule(follow=True))

    def test_validate_default_attributes(self):
        # test null Rule
        rule = Rule(follow=True)
        self.failUnlessEqual(None, rule.matcher)
        self.failUnlessEqual(None, rule.callback)
        self.failUnlessEqual({}, rule.cb_kwargs)
        # follow default False
        self.failUnlessEqual(True, rule.follow)

    def test_validate_attributes_set(self):
        matcher = BaseMatcher()
        callback = lambda: True
        rule = Rule(matcher, callback, True, a=1)
        # test attributes
        self.failUnlessEqual(matcher, rule.matcher)
        self.failUnlessEqual(callback, rule.callback)
        self.failUnlessEqual({'a': 1}, rule.cb_kwargs)
        self.failUnlessEqual(True, rule.follow)

class CompiledRuleInitializationTest(unittest.TestCase):

    def test_fail_on_invalid_matcher(self):
        # pass with valid matcher
        self.failUnless(CompiledRule(BaseMatcher()),
                "Failed CompiledRule instantiation")

        # at least needs valid matcher
        self.assertRaises(AssertionError, CompiledRule, None)
        self.assertRaises(AssertionError, CompiledRule, False)
        self.assertRaises(AssertionError, CompiledRule, True)

    def test_fail_on_invalid_callback(self):
        # pass with valid callback
        callback = lambda: True
        self.failUnless(CompiledRule(BaseMatcher(), callback))
        # pass with callback none
        self.failUnless(CompiledRule(BaseMatcher(), None))

        # assert on invalid callback
        self.assertRaises(AssertionError, CompiledRule, BaseMatcher(),
                          'myfunc')

        # numeric variable
        var = 123
        self.assertRaises(AssertionError, CompiledRule, BaseMatcher(),
                          var)

        class A:
            pass

        # random instance
        self.assertRaises(AssertionError, CompiledRule, BaseMatcher(),
                          A())


    def test_fail_on_invalid_follow_value(self):
        callback = lambda: True
        matcher = BaseMatcher()
        # pass bool
        self.failUnless(CompiledRule(matcher, callback, True))
        self.failUnless(CompiledRule(matcher, callback, False))

        # assert with non-bool
        self.assertRaises(AssertionError, CompiledRule, matcher,
                          callback, None)
        self.assertRaises(AssertionError, CompiledRule, matcher,
                          callback, 1)

    def test_validate_default_attributes(self):
        callback = lambda: True
        matcher = BaseMatcher()
        rule = CompiledRule(matcher, callback, True)

        # test attributes
        self.failUnlessEqual(matcher, rule.matcher)
        self.failUnlessEqual(callback, rule.callback)
        self.failUnlessEqual(True, rule.follow)


class RulesTest(unittest.TestCase):
    def test_rules_manager_basic(self):
        spider = BaseSpider('foo')
        response1 = HtmlResponse('http://example.org')
        response2 = HtmlResponse('http://othersite.org')
        rulesman = RulesManager([], spider)

        # should return none
        self.failIf(rulesman.get_rule_from_response(response1))
        self.failIf(rulesman.get_rule_from_response(response2))

        # rules manager with match-all rule
        rulesman = RulesManager([
                Rule(BaseMatcher(), follow=True),
                ], spider)

        # returns CompiledRule
        rule1 = rulesman.get_rule_from_response(response1)
        rule2 = rulesman.get_rule_from_response(response2)

        self.failUnless(isinstance(rule1, CompiledRule))
        self.failUnless(isinstance(rule2, CompiledRule))
        self.assert_(rule1 is rule2)
        self.failUnlessEqual(rule1.callback, None)
        self.failUnlessEqual(rule1.follow, True)

    def test_rules_manager_empty_rule(self):
        spider = BaseSpider('foo')
        response = HtmlResponse('http://example.org')

        rulesman = RulesManager([Rule(follow=True)], spider)

        rule = rulesman.get_rule_from_response(response)
        # default matcher if None: BaseMatcher
        self.failUnless(isinstance(rule.matcher, BaseMatcher))

    def test_rules_manager_default_matcher(self):
        spider = BaseSpider('foo')
        response = HtmlResponse('http://example.org')
        callback = lambda x: None

        rulesman = RulesManager([
            Rule('http://example.org', callback),
                ], spider, default_matcher=UrlMatcher)

        rule = rulesman.get_rule_from_response(response)
        self.failUnless(isinstance(rule.matcher, UrlMatcher))

    def test_rules_manager_matchers(self):
        spider = BaseSpider('foo')
        response1 = HtmlResponse('http://example.org')
        response2 = HtmlResponse('http://othersite.org')

        urlmatcher = UrlMatcher('http://example.org')
        basematcher = BaseMatcher()
        # callback needed for Rule
        callback = lambda x: None

        # test fail matcher resolve
        self.assertRaises(ValueError, RulesManager,
                          [Rule(False, callback)], spider)
        self.assertRaises(ValueError, RulesManager,
                          [Rule(spider, callback)], spider)

        rulesman = RulesManager([
            Rule(urlmatcher, callback),
            Rule(basematcher, callback),
            ], spider)

        # response1 matches example.org
        rule1 = rulesman.get_rule_from_response(response1)
        # response2 is catch by BaseMatcher()
        rule2 = rulesman.get_rule_from_response(response2)

        self.failUnlessEqual(rule1.matcher, urlmatcher)
        self.failUnlessEqual(rule2.matcher, basematcher)

        # reverse order. BaseMatcher should match all
        rulesman = RulesManager([
            Rule(basematcher, callback),
            Rule(urlmatcher, callback),
            ], spider)

        rule1 = rulesman.get_rule_from_response(response1)
        rule2 = rulesman.get_rule_from_response(response2)

        self.failUnlessEqual(rule1.matcher, basematcher)
        self.failUnlessEqual(rule2.matcher, basematcher)
        self.failUnless(rule1 is rule2)

    def test_rules_manager_callbacks(self):
        mycallback = lambda: True

        spider = BaseSpider('foo')
        spider.parse_item = lambda: True

        response1 = HtmlResponse('http://example.org')
        response2 = HtmlResponse('http://othersite.org')

        rulesman = RulesManager([
            Rule('example', mycallback),
            Rule('othersite', 'parse_item'),
                ], spider, default_matcher=UrlRegexMatcher)

        rule1 = rulesman.get_rule_from_response(response1)
        rule2 = rulesman.get_rule_from_response(response2)

        self.failUnlessEqual(rule1.callback, mycallback)
        self.failUnlessEqual(rule2.callback, spider.parse_item)

        # fail unknown callback
        self.assertRaises(AttributeError, RulesManager, [
                            Rule(BaseMatcher(), 'mycallback')
                            ], spider)
        # fail not callable
        spider.not_callable = True
        self.assertRaises(AttributeError, RulesManager, [
                            Rule(BaseMatcher(), 'not_callable')
                            ], spider)


    def test_rules_manager_callback_with_arguments(self):
        spider = BaseSpider('foo')
        response = HtmlResponse('http://example.org')

        kwargs = {'a': 1}
        
        def myfunc(**mykwargs):
            return mykwargs
        
        # verify return validation
        self.failUnlessEquals(kwargs, myfunc(**kwargs))

        # test callback w/o arguments
        rulesman = RulesManager([
            Rule(BaseMatcher(), myfunc),
            ], spider)
        rule = rulesman.get_rule_from_response(response)

        # without arguments should return same callback
        self.failUnlessEqual(rule.callback, myfunc)

        # test callback w/ arguments
        rulesman = RulesManager([
            Rule(BaseMatcher(), myfunc, **kwargs),
            ], spider)
        rule = rulesman.get_rule_from_response(response)

        # with argument should return partial applied callback
        self.failUnless(isinstance(rule.callback, partial))
        self.failUnlessEquals(kwargs, rule.callback())





from scrapy.command import ScrapyCommand
from scrapy.conf import settings

class Command(ScrapyCommand):

    requires_project = True
    default_settings = {'KEEP_ALIVE': True}

    def short_desc(self):
        return "Start Scrapy in server mode"

    def run(self, args, opts):
        self.crawler.start()



import unittest

from scrapy.http import Request
from scrapy.spider import BaseSpider
from scrapy.contrib.dupefilter import RequestFingerprintDupeFilter, NullDupeFilter


class RequestFingerprintDupeFilterTest(unittest.TestCase):

    def test_filter(self):
        spider = BaseSpider('foo')
        filter = RequestFingerprintDupeFilter()
        filter.open_spider(spider)

        r1 = Request('http://scrapytest.org/1')
        r2 = Request('http://scrapytest.org/2')
        r3 = Request('http://scrapytest.org/2')

        assert not filter.request_seen(spider, r1)
        assert filter.request_seen(spider, r1)

        assert not filter.request_seen(spider, r2)
        assert filter.request_seen(spider, r3)

        filter.close_spider(spider)


class NullDupeFilterTest(unittest.TestCase):

    def test_filter(self):
        spider = BaseSpider('foo')
        filter = NullDupeFilter()
        filter.open_spider(spider)

        r1 = Request('http://scrapytest.org/1')
        assert not filter.request_seen(spider, r1)
        filter.close_spider(spider)



import cgi
import traceback
from cStringIO import StringIO

from twisted.web.resource import Resource

from scrapy.utils.txweb import JsonResource
from .interfaces import IPoller, IEggStorage, ISpiderScheduler
from .eggutils import get_spider_list_from_eggfile

class WsResource(JsonResource):

    def __init__(self, root):
        JsonResource.__init__(self)
        self.root = root

    def render(self, txrequest):
        try:
            return JsonResource.render(self, txrequest)
        except Exception, e:
            if self.root.debug:
                return traceback.format_exc()
            r = {"status": "error", "message": str(e)}
            return self.render_object(r, txrequest)

class Schedule(WsResource):

    def render_POST(self, txrequest):
        args = dict((k, v[0]) for k, v in txrequest.args.items())
        project = args.pop('project')
        spider = args.pop('spider')
        sched = self.root.app.getComponent(ISpiderScheduler)
        sched.schedule(project, spider, **args)
        return {"status": "ok"}

class AddVersion(WsResource):

    def render_POST(self, txrequest):
        ct = txrequest.requestHeaders.getRawHeaders('Content-Type')[0]
        boundary = ct.split('boundary=', 1)[1]
        d = cgi.parse_multipart(txrequest.content, {'boundary': boundary})
        project = d['project'][0]
        version = d['version'][0]
        eggf = StringIO(d['egg'][0])
        spiders = get_spider_list_from_eggfile(eggf, project)
        eggstorage = self.root.app.getComponent(IEggStorage)
        eggstorage.put(eggf, project, version)
        self.root.update_projects()
        return {"status": "ok", "spiders": spiders}

class ListProjects(WsResource):

    def render_GET(self, txrequest):
        projects = self.root.app.getComponent(ISpiderScheduler).list_projects()
        return {"status": "ok", "projects": projects}

class ListVersions(WsResource):

    def render_GET(self, txrequest):
        project = txrequest.args['project'][0]
        eggstorage = self.root.app.getComponent(IEggStorage)
        versions = eggstorage.list(project)
        return {"status": "ok", "versions": versions}

class ListSpiders(WsResource):

    def render_GET(self, txrequest):
        project = txrequest.args['project'][0]
        eggstorage = self.root.app.getComponent(IEggStorage)
        _, eggf = eggstorage.get(project)
        spiders = get_spider_list_from_eggfile(eggf, project)
        return {"status": "ok", "spiders": spiders}

class DeleteProject(WsResource):

    def render_POST(self, txrequest):
        project = txrequest.args['project'][0]
        self._delete_version(project)
        return {"status": "ok"}

    def _delete_version(self, project, version=None):
        eggstorage = self.root.app.getComponent(IEggStorage)
        eggstorage.delete(project, version)
        self.root.update_projects()

class DeleteVersion(DeleteProject):

    def render_POST(self, txrequest):
        project = txrequest.args['project'][0]
        version = txrequest.args['version'][0]
        self._delete_version(project, version)
        return {"status": "ok"}

class Root(Resource):

    def __init__(self, config, app):
        Resource.__init__(self)
        self.debug = config.getboolean('debug', False)
        self.app = app
        self.putChild('schedule.json', Schedule(self))
        self.putChild('addversion.json', AddVersion(self))
        self.putChild('listprojects.json', ListProjects(self))
        self.putChild('listversions.json', ListVersions(self))
        self.putChild('listspiders.json', ListSpiders(self))
        self.putChild('delproject.json', DeleteProject(self))
        self.putChild('delversion.json', DeleteVersion(self))
        self.update_projects()

    def update_projects(self):
        self.app.getComponent(IPoller).update_projects()
        self.app.getComponent(ISpiderScheduler).update_projects()



"""
An extension to retry failed requests that are potentially caused by temporary
problems such as a connection timeout or HTTP 500 error.

You can change the behaviour of this middleware by modifing the scraping settings:
RETRY_TIMES - how many times to retry a failed page
RETRY_HTTP_CODES - which HTTP response codes to retry

Failed pages are collected on the scraping process and rescheduled at the end,
once the spider has finished crawling all regular (non failed) pages. Once
there is no more failed pages to retry this middleware sends a signal
(retry_complete), so other extensions could connect to that signal.

Default values are located in scrapy.conf.default_settings, like any other
setting

About HTTP errors to consider:

- You may want to remove 400 from RETRY_HTTP_CODES, if you stick to the HTTP
  protocol. It's included by default because it's a common code used to
  indicate server overload, which would be something we want to retry
"""

from twisted.internet.error import TimeoutError as ServerTimeoutError, DNSLookupError, \
                                   ConnectionRefusedError, ConnectionDone, ConnectError, \
                                   ConnectionLost
from twisted.internet.defer import TimeoutError as UserTimeoutError
from twisted.web.client import PartialDownloadError

from scrapy import log
from scrapy.utils.response import response_status_message
from scrapy.conf import settings

class RetryMiddleware(object):

    EXCEPTIONS_TO_RETRY = (ServerTimeoutError, UserTimeoutError, DNSLookupError,
                           ConnectionRefusedError, ConnectionDone, ConnectError,
                           ConnectionLost, PartialDownloadError)

    def __init__(self):
        self.max_retry_times = settings.getint('RETRY_TIMES')
        self.retry_http_codes = map(int, settings.getlist('RETRY_HTTP_CODES'))
        self.priority_adjust = settings.getint('RETRY_PRIORITY_ADJUST')

    def process_response(self, request, response, spider):
        if 'dont_retry' in request.meta:
            return response
        if response.status in self.retry_http_codes:
            reason = response_status_message(response.status)
            return self._retry(request, reason, spider) or response
        return response

    def process_exception(self, request, exception, spider):
        if isinstance(exception, self.EXCEPTIONS_TO_RETRY) \
                and 'dont_retry' not in request.meta:
            return self._retry(request, exception, spider)

    def _retry(self, request, reason, spider):
        retries = request.meta.get('retry_times', 0) + 1

        if retries <= self.max_retry_times:
            log.msg("Retrying %s (failed %d times): %s" % (request, retries, reason),
                    spider=spider, level=log.DEBUG)
            retryreq = request.copy()
            retryreq.meta['retry_times'] = retries
            retryreq.dont_filter = True
            retryreq.priority = request.priority + self.priority_adjust
            return retryreq
        else:
            log.msg("Discarding %s (failed %d times): %s" % (request, retries, reason),
                    spider=spider, level=log.DEBUG)




import sys, os
from shutil import copyfileobj
from tempfile import mkstemp

from twisted.internet import reactor, defer, protocol, error
from twisted.application.service import Service
from twisted.python import log

from scrapy.utils.py26 import cpu_count
from .interfaces import IPoller, IEggStorage, IEnvironment

class Launcher(Service):

    def __init__(self, config, app):
        self.max_proc = config.getint('max_proc', 0) or cpu_count()
        self.egg_runner = config.get('egg_runner', 'scrapyd.eggrunner')
        self.app = app

    def startService(self):
        for slot in range(self.max_proc):
            self._wait_for_project(slot)
        log.msg("%s started: max_proc=%r, egg_runner=%r" % (self.parent.name, \
            self.max_proc, self.egg_runner), system="Launcher")

    def _wait_for_project(self, slot):
        poller = self.app.getComponent(IPoller)
        poller.next().addCallback(self._spawn_process, slot)

    def _get_eggpath(self, project):
        eggstorage = self.app.getComponent(IEggStorage)
        version, eggf = eggstorage.get(project)
        if eggf is None:
            return
        prefix = '%s-%s-' % (project, version)
        fd, eggpath = mkstemp(prefix=prefix, suffix='.egg')
        lf = os.fdopen(fd, 'wb')
        copyfileobj(eggf, lf)
        lf.close()
        return eggpath

    def _spawn_process(self, message, slot):
        project = message['project']
        eggpath = self._get_eggpath(project)
        args = [sys.executable, '-m', self.egg_runner, 'crawl']
        e = self.app.getComponent(IEnvironment)
        env = e.get_environment(message, slot, eggpath)
        pp = ScrapyProcessProtocol(eggpath, slot)
        pp.deferred.addBoth(self._process_finished, eggpath, slot)
        reactor.spawnProcess(pp, sys.executable, args=args, env=env)

    def _process_finished(self, _, eggpath, slot):
        if eggpath:
            os.remove(eggpath)
        self._wait_for_project(slot)


class ScrapyProcessProtocol(protocol.ProcessProtocol):

    def __init__(self, eggfile, slot):
        self.eggfile = eggfile
        self.slot = slot
        self.pid = None
        self.deferred = defer.Deferred()

    def outReceived(self, data):
        log.msg(data.rstrip(), system="Launcher,%d/stdout" % self.pid)

    def errReceived(self, data):
        log.msg(data.rstrip(), system="Launcher,%d/stderr" % self.pid)

    def connectionMade(self):
        self.pid = self.transport.pid
        self.log("Process started: ")

    def processEnded(self, status):
        if isinstance(status.value, error.ProcessDone):
            self.log("Process finished: ")
        else:
            self.log("Process died: exitstatus=%r " % status.value.exitCode)
        self.deferred.callback(self)

    def log(self, msg):
        msg += "slot=%r pid=%r egg=%r" % (self.slot, self.pid, self.eggfile)
        log.msg(msg, system="Launcher")



"""
Scrapy engine tests

This starts a testing web server (using twisted.server.Site) and then crawls it
with the Scrapy crawler.

To view the testing web server in a browser you can start it by running this
module with the ``runserver`` argument::

    python test_engine.py runserver
"""

import sys, os, re, urlparse

from twisted.internet import reactor, defer
from twisted.web import server, static, util
from twisted.trial import unittest

from scrapy import signals
from scrapy.conf import Settings
from scrapy.crawler import Crawler
from scrapy.xlib.pydispatch import dispatcher
from scrapy.tests import tests_datadir
from scrapy.spider import BaseSpider
from scrapy.item import Item, Field
from scrapy.contrib.linkextractors.sgml import SgmlLinkExtractor
from scrapy.http import Request
from scrapy.utils.signal import disconnect_all

class TestItem(Item):
    name = Field()
    url = Field()
    price = Field()

class TestSpider(BaseSpider):
    name = "scrapytest.org"
    allowed_domains = ["scrapytest.org", "localhost"]

    itemurl_re = re.compile("item\d+.html")
    name_re = re.compile("<h1>(.*?)</h1>", re.M)
    price_re = re.compile(">Price: \$(.*?)<", re.M)

    def parse(self, response):
        xlink = SgmlLinkExtractor()
        itemre = re.compile(self.itemurl_re)
        for link in xlink.extract_links(response):
            if itemre.search(link.url):
                yield Request(url=link.url, callback=self.parse_item)

    def parse_item(self, response):
        item = TestItem()
        m = self.name_re.search(response.body)
        if m:
            item['name'] = m.group(1)
        item['url'] = response.url
        m = self.price_re.search(response.body)
        if m:
            item['price'] = m.group(1)
        return item

def start_test_site(debug=False):
    root_dir = os.path.join(tests_datadir, "test_site")
    r = static.File(root_dir)
    r.putChild("redirect", util.Redirect("/redirected"))
    r.putChild("redirected", static.Data("Redirected here", "text/plain"))

    port = reactor.listenTCP(0, server.Site(r), interface="127.0.0.1")
    if debug:
        print "Test server running at http://localhost:%d/ - hit Ctrl-C to finish." \
            % port.getHost().port
    return port


class CrawlerRun(object):
    """A class to run the crawler and keep track of events occurred"""

    def __init__(self):
        self.spider = None
        self.respplug = []
        self.reqplug = []
        self.itemresp = []
        self.signals_catched = {}

    def run(self):
        self.port = start_test_site()
        self.portno = self.port.getHost().port

        start_urls = [self.geturl("/"), self.geturl("/redirect")]
        self.spider = TestSpider(start_urls=start_urls)

        for name, signal in vars(signals).items():
            if not name.startswith('_'):
                dispatcher.connect(self.record_signal, signal)
        dispatcher.connect(self.item_scraped, signals.item_scraped)
        dispatcher.connect(self.request_received, signals.request_received)
        dispatcher.connect(self.response_downloaded, signals.response_downloaded)

        settings = Settings()
        self.crawler = Crawler(settings)
        self.crawler.install()
        self.crawler.configure()
        self.crawler.queue.append_spider(self.spider)
        self.crawler.start()

        self.deferred = defer.Deferred()
        dispatcher.connect(self.stop, signals.engine_stopped)
        return self.deferred

    def stop(self):
        self.port.stopListening()
        for name, signal in vars(signals).items():
            if not name.startswith('_'):
                disconnect_all(signal)
        self.crawler.uninstall()
        self.deferred.callback(None)

    def geturl(self, path):
        return "http://localhost:%s%s" % (self.portno, path)

    def getpath(self, url):
        u = urlparse.urlparse(url)
        return u.path

    def item_scraped(self, item, spider, response):
        self.itemresp.append((item, response))

    def request_received(self, request, spider):
        self.reqplug.append((request, spider))

    def response_downloaded(self, response, spider):
        self.respplug.append((response, spider))

    def record_signal(self, *args, **kwargs):
        """Record a signal and its parameters"""
        signalargs = kwargs.copy()
        sig = signalargs.pop('signal')
        signalargs.pop('sender', None)
        self.signals_catched[sig] = signalargs


class EngineTest(unittest.TestCase):

    @defer.inlineCallbacks
    def test_crawler(self):
        self.run = CrawlerRun()
        yield self.run.run()
        self._assert_visited_urls()
        self._assert_received_requests()
        self._assert_downloaded_responses()
        self._assert_scraped_items()
        self._assert_signals_catched()

    def _assert_visited_urls(self):
        must_be_visited = ["/", "/redirect", "/redirected", 
                           "/item1.html", "/item2.html", "/item999.html"]
        urls_visited = set([rp[0].url for rp in self.run.respplug])
        urls_expected = set([self.run.geturl(p) for p in must_be_visited])
        assert urls_expected <= urls_visited, "URLs not visited: %s" % list(urls_expected - urls_visited)

    def _assert_received_requests(self):
        # 3 requests should be received from the spider. start_urls and redirects don't count
        self.assertEqual(3, len(self.run.reqplug))

        paths_expected = ['/item999.html', '/item2.html', '/item1.html']

        urls_requested = set([rq[0].url for rq in self.run.reqplug])
        urls_expected = set([self.run.geturl(p) for p in paths_expected])
        assert urls_expected <= urls_requested

    def _assert_downloaded_responses(self):
        # response tests
        self.assertEqual(6, len(self.run.respplug))

        for response, _ in self.run.respplug:
            if self.run.getpath(response.url) == '/item999.html':
                self.assertEqual(404, response.status)
            if self.run.getpath(response.url) == '/redirect':
                self.assertEqual(302, response.status)

    def _assert_scraped_items(self):
        self.assertEqual(2, len(self.run.itemresp))
        for item, response in self.run.itemresp:
            self.assertEqual(item['url'], response.url)
            if 'item1.html' in item['url']:
                self.assertEqual('Item 1 name', item['name'])
                self.assertEqual('100', item['price'])
            if 'item2.html' in item['url']:
                self.assertEqual('Item 2 name', item['name'])
                self.assertEqual('200', item['price'])

    def _assert_signals_catched(self):
        assert signals.engine_started in self.run.signals_catched
        assert signals.engine_stopped in self.run.signals_catched
        assert signals.spider_opened in self.run.signals_catched
        assert signals.spider_idle in self.run.signals_catched
        assert signals.spider_closed in self.run.signals_catched

        self.assertEqual({'spider': self.run.spider},
                         self.run.signals_catched[signals.spider_opened])
        self.assertEqual({'spider': self.run.spider},
                         self.run.signals_catched[signals.spider_idle])
        self.assertEqual({'spider': self.run.spider, 'reason': 'finished'},
                         self.run.signals_catched[signals.spider_closed])


if __name__ == "__main__":
    if len(sys.argv) > 1 and sys.argv[1] == 'runserver':
        start_test_site(debug=True)
        reactor.run()



from zope.interface import Interface

class ISpiderQueue(Interface):

    def from_settings(settings):
        """Class method to instantiate from settings"""

    def add(name, **spider_args):
        """Add a spider to the queue given its name a some spider arguments.
        
        This method can return a deferred. """

    def pop():
        """Pop the next mesasge from the queue. The messages is a dict
        conaining a key 'name' with the spider name and other keys as spider
        attributes.

        This method can return a deferred. """

    def list():
        """Return a list with the messages in the queue. Each message is a dict
        which must have a 'name' key (with the spider name), and other optional
        keys that will be used as spider arguments, to create the spider.
        
        This method can return a deferred. """

    def count():
        """Return the number of spiders in the queue.
        
        This method can return a deferred. """

    def clear():
        """Clear the queue.

        This method can return a deferred. """

class ISpiderManager(Interface):

    def from_settings(settings):
        """Class method to instantiate from settings"""

    def create(spider_name, **spider_args):
        """Returns a new Spider instance for the given spider name, and using
        the given spider arguments. If the spider name is not found, it must
        raise a KeyError."""

    def list():
        """Return a list with the names of all spiders available in the
        project"""

    def find_by_request(request):
        """Returns the list of spiders names that can handle the given request"""



from twisted.trial import unittest
from twisted.python import failure
from twisted.internet import defer, reactor

from scrapy.conf import Settings
from scrapy.crawler import Crawler
from scrapy.http import Request, Response
from scrapy.spider import BaseSpider
from scrapy.utils.request import request_fingerprint
from scrapy.contrib.pipeline.media import MediaPipeline


class _MockedMediaPipeline(MediaPipeline):

    def download(self, request, info):
        delay = request.meta.get('delay')
        response = request.meta.get('response')
        if delay is None:
            return response
        else:
            dfd = defer.Deferred()
            reactor.callLater(delay, dfd.callback, None)
            return dfd.addCallback(lambda _: response)

    def get_media_requests(self, item, info):
        return item.get('requests')


class MediaPipelineTestCase(unittest.TestCase):

    pipeline_class = _MockedMediaPipeline

    def setUp(self):
        self.crawler = Crawler(Settings())
        self.crawler.install()
        self.spider = BaseSpider('media.com')
        self.pipe = self.pipeline_class()
        self.pipe.open_spider(self.spider)

    def tearDown(self):
        self.pipe.close_spider(self.spider)
        self.crawler.uninstall()

    @defer.inlineCallbacks
    def test_return_item_by_default(self):
        item = dict(name='sofa')
        new_item = yield self.pipe.process_item(item, self.spider)
        assert new_item is item

    @defer.inlineCallbacks
    def test_get_media_requests(self):
        # returns single Request (without callback)
        info = self.pipe.spiderinfo[self.spider]
        req = Request('http://media.com/2.gif')
        item = dict(requests=req) # pass a single item
        new_item = yield self.pipe.process_item(item, self.spider)
        assert new_item is item
        assert request_fingerprint(req) in info.downloaded

        # returns iterable of Requests
        req1 = Request('http://media.com/1.gif')
        req2 = Request('http://media.com/1.jpg')
        item = dict(requests=iter([req1, req2]))
        new_item = yield self.pipe.process_item(item, self.spider)
        assert new_item is item
        assert info.downloaded.get(request_fingerprint(req1)) is None
        assert info.downloaded.get(request_fingerprint(req2)) is None

    @defer.inlineCallbacks
    def test_requets_callback_is_called(self):
        collected = []
        response = Response('http://media.com/2.gif')
        request = Request('http://media.com/2.gif', meta=dict(response=response), callback=collected.append)
        item = dict(requests=request) # pass a single item
        yield self.pipe.process_item(item, self.spider)
        assert collected == [response]

    @defer.inlineCallbacks
    def test_requets_errback_is_called(self):
        collected = []
        fail = failure.Failure(Exception())
        req = Request('http://media.com/2.gif', meta=dict(response=fail), callback=lambda _:_, errback=collected.append)
        item = dict(requests=req) # pass a single item
        yield self.pipe.process_item(item, self.spider)
        assert collected == [fail]



"""This module implements the Scraper component which parses responses and
extracts information from them"""

from twisted.python.failure import Failure
from twisted.internet import defer

from scrapy.utils.defer import defer_result, defer_succeed, parallel, iter_errback
from scrapy.utils.spider import iterate_spider_output
from scrapy.utils.misc import load_object
from scrapy.utils.signal import send_catch_log, send_catch_log_deferred
from scrapy.exceptions import IgnoreRequest, DropItem
from scrapy import signals
from scrapy.http import Request, Response
from scrapy.item import BaseItem
from scrapy.core.spidermw import SpiderMiddlewareManager
from scrapy import log
from scrapy.stats import stats


class SpiderInfo(object):
    """Object for holding data of the responses being scraped"""

    MIN_RESPONSE_SIZE = 1024

    def __init__(self, max_active_size=5000000):
        self.max_active_size = max_active_size
        self.queue = []
        self.active = set()
        self.active_size = 0
        self.itemproc_size = 0
        self.closing = None

    def add_response_request(self, response, request):
        deferred = defer.Deferred()
        self.queue.append((response, request, deferred))
        if isinstance(response, Response):
            self.active_size += max(len(response.body), self.MIN_RESPONSE_SIZE)
        else:
            self.active_size += self.MIN_RESPONSE_SIZE
        return deferred

    def next_response_request_deferred(self):
        response, request, deferred = self.queue.pop(0)
        self.active.add(response)
        return response, request, deferred

    def finish_response(self, response):
        self.active.remove(response)
        if isinstance(response, Response):
            self.active_size -= max(len(response.body), self.MIN_RESPONSE_SIZE)
        else:
            self.active_size -= self.MIN_RESPONSE_SIZE

    def is_idle(self):
        return not (self.queue or self.active)

    def needs_backout(self):
        return self.active_size > self.max_active_size

class Scraper(object):

    def __init__(self, engine, settings):
        self.sites = {}
        self.spidermw = SpiderMiddlewareManager.from_settings(settings)
        itemproc_cls = load_object(settings['ITEM_PROCESSOR'])
        self.itemproc = itemproc_cls.from_settings(settings)
        self.concurrent_items = settings.getint('CONCURRENT_ITEMS')
        self.engine = engine

    @defer.inlineCallbacks
    def open_spider(self, spider):
        """Open the given spider for scraping and allocate resources for it"""
        assert spider not in self.sites, "Spider already opened: %s" % spider
        self.sites[spider] = SpiderInfo()
        yield self.itemproc.open_spider(spider)

    def close_spider(self, spider):
        """Close a spider being scraped and release its resources"""
        assert spider in self.sites, "Spider not opened: %s" % spider
        site = self.sites[spider]
        site.closing = defer.Deferred()
        site.closing.addCallback(self.itemproc.close_spider)
        self._check_if_closing(spider, site)
        return site.closing

    def is_idle(self):
        """Return True if there isn't any more spiders to process"""
        return not self.sites

    def _check_if_closing(self, spider, site):
        if site.closing and site.is_idle():
            del self.sites[spider]
            site.closing.callback(spider)

    def enqueue_scrape(self, response, request, spider):
        site = self.sites.get(spider, None)
        if site is None:
            return
        dfd = site.add_response_request(response, request)
        def finish_scraping(_):
            site.finish_response(response)
            self._check_if_closing(spider, site)
            self._scrape_next(spider, site)
            return _
        dfd.addBoth(finish_scraping)
        dfd.addErrback(log.err, 'Scraper bug processing %s' % request, \
            spider=spider)
        self._scrape_next(spider, site)
        return dfd

    def _scrape_next(self, spider, site):
        while site.queue:
            response, request, deferred = site.next_response_request_deferred()
            self._scrape(response, request, spider).chainDeferred(deferred)

    def _scrape(self, response, request, spider):
        """Handle the downloaded response or failure trough the spider
        callback/errback"""
        assert isinstance(response, (Response, Failure))

        dfd = self._scrape2(response, request, spider) # returns spiders processed output
        dfd.addErrback(self.handle_spider_error, request, spider)
        dfd.addCallback(self.handle_spider_output, request, response, spider)
        return dfd

    def _scrape2(self, request_result, request, spider):
        """Handle the diferent cases of request's result been a Response or a
        Failure"""
        if not isinstance(request_result, Failure):
            return self.spidermw.scrape_response(self.call_spider, \
                request_result, request, spider)
        else:
            # FIXME: don't ignore errors in spider middleware
            dfd = self.call_spider(request_result, request, spider)
            return dfd.addErrback(self._check_propagated_failure, \
                request_result, request, spider)

    def call_spider(self, result, request, spider):
        dfd = defer_result(result)
        dfd.addCallbacks(request.callback or spider.parse, request.errback)
        return dfd.addCallback(iterate_spider_output)

    def handle_spider_error(self, _failure, request, spider, propagated_failure=None):
        referer = request.headers.get('Referer', None)
        msg = "Spider error processing <%s> (referer: <%s>)" % \
            (request.url, referer)
        log.err(_failure, msg, spider=spider)
        stats.inc_value("spider_exceptions/%s" % _failure.value.__class__.__name__, \
            spider=spider)

    def handle_spider_output(self, result, request, response, spider):
        if not result:
            return defer_succeed(None)
        it = iter_errback(result, self.handle_spider_error, request, spider)
        dfd = parallel(it, self.concurrent_items,
            self._process_spidermw_output, request, response, spider)
        return dfd

    def _process_spidermw_output(self, output, request, response, spider):
        """Process each Request/Item (given in the output parameter) returned
        from the given spider
        """
        if isinstance(output, Request):
            send_catch_log(signal=signals.request_received, request=output, \
                spider=spider)
            self.engine.crawl(request=output, spider=spider)
        elif isinstance(output, BaseItem):
            log.msg(log.formatter.scraped(output, request, response, spider), \
                level=log.DEBUG, spider=spider)
            self.sites[spider].itemproc_size += 1
            dfd = send_catch_log_deferred(signal=signals.item_scraped, \
                item=output, spider=spider, response=response)
            dfd.addBoth(lambda _: self.itemproc.process_item(output, spider))
            dfd.addBoth(self._itemproc_finished, output, spider)
            return dfd
        elif output is None:
            pass
        else:
            log.msg("Spider must return Request, BaseItem or None, got %r in %s" % \
                (type(output).__name__, request), log.ERROR, spider=spider)

    def _check_propagated_failure(self, spider_failure, propagated_failure, request, spider):
        """Log and silence the bugs raised outside of spiders, but still allow
        spiders to be notified about general failures while downloading spider
        generated requests
        """
        # ignored requests are commonly propagated exceptions safes to be silenced
        if isinstance(spider_failure.value, IgnoreRequest):
            return
        elif spider_failure is propagated_failure:
            log.err(spider_failure, 'Unhandled error propagated to spider', \
                spider=spider)
            return # stop propagating this error
        else:
            return spider_failure # exceptions raised in the spider code

    def _itemproc_finished(self, output, item, spider):
        """ItemProcessor finished for the given ``item`` and returned ``output``
        """
        self.sites[spider].itemproc_size -= 1
        if isinstance(output, Failure):
            ex = output.value
            if isinstance(ex, DropItem):
                log.msg(log.formatter.dropped(item, ex, spider), \
                    level=log.WARNING, spider=spider)
                return send_catch_log_deferred(signal=signals.item_dropped, \
                    item=item, spider=spider, exception=output.value)
            else:
                log.err(output, 'Error processing %s' % item, spider=spider)
        else:
            log.msg(log.formatter.passed(output, spider), log.INFO, spider=spider)
            return send_catch_log_deferred(signal=signals.item_passed, \
                item=item, spider=spider, output=output)




"""
htmlpage

Container object for representing html pages in the IBL system. This
encapsulates page related information and prevents parsing multiple times.
"""
import re
import hashlib

from scrapy.utils.python import str_to_unicode

def create_page_from_jsonpage(jsonpage, body_key):
    """Create an HtmlPage object from a dict object conforming to the schema
    for a page

    `body_key` is the key where the body is stored and can be either 'body'
    (original page with annotations - if any) or 'original_body' (original
    page, always). Classification typically uses 'original_body' to avoid
    confusing the classifier with annotated pages, while extraction uses 'body'
    to pass the annotated pages.
    """
    url = jsonpage['url']
    headers = jsonpage.get('headers')
    body = str_to_unicode(jsonpage[body_key])
    page_id = jsonpage.get('page_id')
    return HtmlPage(url, headers, body, page_id)

class HtmlPage(object):
    def __init__(self, url=None, headers=None, body=None, page_id=None):
        assert isinstance(body, unicode), "unicode expected, got: %s" % type(body).__name__
        self.headers = headers or {}
        self.body = body
        self.url = url or u''
        if page_id is None and url:
            self.page_id = hashlib.sha1(url).hexdigest()
        else:
            self.page_id = page_id 
    

    def _set_body(self, body):
        self._body = body
        self.parsed_body = list(parse_html(body))
        
    body = property(lambda x: x._body, _set_body)
    
    def fragment_data(self, data_fragment):
        return self.body[data_fragment.start:data_fragment.end]
    
class HtmlTagType(object):
    OPEN_TAG = 1
    CLOSE_TAG = 2 
    UNPAIRED_TAG = 3

class HtmlDataFragment(object):
    __slots__ = ('start', 'end')
    
    def __init__(self, start, end):
        self.start = start
        self.end = end
        
    def __str__(self):
        return "<HtmlDataFragment [%s:%s]>" % (self.start, self.end)

    def __repr__(self):
        return str(self)
    
class HtmlTag(HtmlDataFragment):
    __slots__ = ('tag_type', 'tag', 'attributes')

    def __init__(self, tag_type, tag, attributes, start, end):
        HtmlDataFragment.__init__(self, start, end)
        self.tag_type = tag_type
        self.tag = tag
        self.attributes = attributes

    def __str__(self):
        return "<HtmlTag tag='%s' attributes={%s} [%s:%s]>" % (self.tag, ', '.join(sorted\
                (["%s: %s" % (k, repr(v)) for k, v in self.attributes.items()])), self.start, self.end)
    
    def __repr__(self):
        return str(self)

_ATTR = "((?:[^=/>\s]|/(?!>))+)(?:\s*=(?:\s*\"(.*?)\"|\s*'(.*?)'|([^>\s]+))?)?"
_TAG = "<(\/?)(\w+(?::\w+)?)((?:\s+" + _ATTR + ")+\s*|\s*)(\/?)>"
_DOCTYPE = r"<!DOCTYPE.*?>"

_ATTR_REGEXP = re.compile(_ATTR, re.I | re.DOTALL)
_HTML_REGEXP = re.compile(_TAG, re.I | re.DOTALL)
_DOCTYPE_REGEXP = re.compile("(?:%s)" % _DOCTYPE)
_COMMENT_RE = re.compile("(<!--.*?-->)", re.DOTALL)
_SCRIPT_RE = re.compile("(<script.*?>).*?(</script.*?>)", re.DOTALL | re.I)

def parse_html(text):
    """Higher level html parser. Calls lower level parsers and joins sucesive
    HtmlDataFragment elements in a single one.
    """
    script_layer = lambda x: _parse_clean_html(x, _SCRIPT_RE, HtmlTag, _simple_parse_html)
    comment_layer = lambda x: _parse_clean_html(x, _COMMENT_RE, HtmlDataFragment, script_layer)
    delayed_element = None
    for element in comment_layer(text):
        if isinstance(element, HtmlTag):
            if delayed_element is not None:
                yield delayed_element
                delayed_element = None
            yield element
        else:# element is HtmlDataFragment
            if delayed_element is not None:
                delayed_element.start = min(element.start, delayed_element.start)
                delayed_element.end = max(element.end, delayed_element.end)
            else:
                delayed_element = element
    if delayed_element is not None:
        yield delayed_element

def _parse_clean_html(text, regex, htype, func):
    """
    Removes regions from text, passes the cleaned text to the lower parse layer,
    and reinserts removed regions.
    regex - regular expression that defines regions to be removed/re inserted
    htype - the html parser type of the removed elements
    func - function that performs the lower parse layer
    """
    removed = [[m.start(), m.end(), m.groups()] for m in regex.finditer(text)]
    
    cleaned = regex.sub("", text)
    shift = 0
    for element in func(cleaned):
        element.start += shift
        element.end += shift
        while removed:
            if element.end <= removed[0][0]:
                yield element
                break
            else:
                start, end, groups = removed.pop(0)
                add = end - start
                element.end += add
                shift += add
                if element.start >= start:
                    element.start += add
                elif isinstance(element, HtmlTag):
                    yield element
                    break
                
                if element.start < start:
                    yield HtmlDataFragment(element.start, start)
                    element.start = end
                
                if htype == HtmlTag:
                    begintag = _parse_tag(_HTML_REGEXP.match(groups[0]))
                    endtag = _parse_tag(_HTML_REGEXP.match(groups[1]))
                    begintag.start = start
                    begintag.end += start
                    
                    endtag.start = end - endtag.end
                    endtag.end = end
                    content = None
                    if begintag.end < endtag.start:
                        content = HtmlDataFragment(begintag.end, endtag.start)
                    yield begintag
                    if content is not None:
                        yield content
                    yield endtag
                else:
                    yield htype(start, end)
        else:
            yield element

def _simple_parse_html(text):
    """Simple html parse. It returns a sequence of HtmlTag and HtmlDataFragment
    objects. Does not ignore any region.
    """
    # If have doctype remove it.
    start_pos = 0
    match = _DOCTYPE_REGEXP.match(text)
    if match:
        start_pos = match.end()
    prev_end = start_pos
    for match in _HTML_REGEXP.finditer(text, start_pos):
        start = match.start()
        end = match.end()
            
        if start > prev_end:
            yield HtmlDataFragment(prev_end, start)
        
        yield _parse_tag(match)
        prev_end = end
    textlen = len(text)
    if prev_end < textlen:
        yield HtmlDataFragment(prev_end, textlen)

def _parse_tag(match):
    """
    parse a tag matched by _HTML_REGEXP
    """
    data = match.groups()
    closing, tag, attr_text = data[:3]
    # if tag is None then the match is a comment
    if tag is not None:
        unpaired = data[-1]
        if closing:
            tag_type = HtmlTagType.CLOSE_TAG
        elif unpaired:
            tag_type = HtmlTagType.UNPAIRED_TAG
        else:
            tag_type = HtmlTagType.OPEN_TAG
        attributes = []
        for attr_match in _ATTR_REGEXP.findall(attr_text):
            name = attr_match[0].lower()
            values = [v for v in attr_match[1:] if v]
            attributes.append((name, values[0] if values else None))
        return HtmlTag(tag_type, tag.lower(), dict(attributes), match.start(), match.end())



"""
Page objects

This module contains objects representing pages and parts of pages (e.g. tokens
and annotations) used in the instance based learning algorithm.
"""
from numpy import array, ndarray

from scrapy.contrib.ibl.htmlpage import HtmlTagType

class TokenType(object):
    """constants for token types"""
    WORD = 0
    OPEN_TAG = HtmlTagType.OPEN_TAG
    CLOSE_TAG = HtmlTagType.CLOSE_TAG
    NON_PAIRED_TAG = HtmlTagType.UNPAIRED_TAG

class TokenDict(object):
    """Mapping from parse tokens to integers
    
    >>> d = TokenDict()
    >>> d.tokenid('i')
    0
    >>> d.tokenid('b')
    1
    >>> d.tokenid('i')
    0

    Tokens can be searched for by id
    >>> d.find_token(1)
    'b'

    The lower 24 bits store the token reference and the higher bits the type.
    """
    
    def __init__(self):
        self.token_ids = {}

    def tokenid(self, token, token_type=TokenType.WORD):
        """create an integer id from the token and token type passed"""
        tid = self.token_ids.setdefault(token, len(self.token_ids))
        return tid | (token_type << 24)
    
    @staticmethod
    def token_type(token):
        """extract the token type from the token id passed"""
        return token >> 24

    def find_token(self, tid):
        """Search for a tag with the given ID

        This is O(N) and is only intended for debugging
        """
        tid &= 0xFFFFFF
        if tid >= len(self.token_ids) or tid < 0:
            raise ValueError("tag id %s out of range" % tid)

        for (token, token_id) in self.token_ids.items():
            if token_id == tid:
                return token
        assert False, "token dictionary is corrupt"

    def token_string(self, tid):
        """create a string representation of a token

        This is O(N).
        """
        templates = ["%s", "<%s>", "</%s>", "<%s/>"]
        return templates[tid >> 24] % self.find_token(tid)

class Page(object):
    """Basic representation of a page. This consists of a reference to a
    dictionary of tokens and an array of raw token ids
    """

    __slots__ = ('token_dict', 'page_tokens')

    def __init__(self, token_dict, page_tokens):
        self.token_dict = token_dict
        # use a numpy array becuase we can index/slice easily and efficiently
        if not isinstance(page_tokens, ndarray):
            page_tokens = array(page_tokens)
        self.page_tokens = page_tokens

class TemplatePage(Page):
    __slots__ = ('annotations', 'id', 'ignored_regions', 'extra_required_attrs')

    def __init__(self, token_dict, page_tokens, annotations, template_id=None, \
            ignored_regions=None, extra_required=None):
        Page.__init__(self, token_dict, page_tokens)
        # ensure order is the same as start tag order in the original page
        annotations = sorted(annotations, key=lambda x: x.end_index, reverse=True)
        self.annotations = sorted(annotations, key=lambda x: x.start_index)
        self.id = template_id
        self.ignored_regions = ignored_regions or []
        self.extra_required_attrs = set(extra_required or [])

    def __str__(self):
        summary = []
        for index, token in enumerate(self.page_tokens):
            text = "%s: %s" % (index, self.token_dict.find_token(token))
            summary.append(text)
        return "TemplatePage\n============\nTokens: (index, token)\n%s\nAnnotations: %s\n" % \
                ('\n'.join(summary), '\n'.join(map(str, self.annotations)))

class ExtractionPage(Page):
    """Parsed data belonging to a web page upon which we wish to perform
    extraction.
    """
    __slots__ = ('text', 
            'token_start_indexes', # index in text of the start of a token
            'token_follow_indexes', # index in text of data following token
            'tag_attributes'        # a map from token index to tag attributes
        )

    def __init__(self, text, token_dict, page_tokens, token_start_indexes, 
            token_follow_indexes, tag_attributes):
        Page.__init__(self, token_dict, page_tokens)
        self.text = text
        self.token_start_indexes = token_start_indexes
        self.token_follow_indexes = token_follow_indexes
        self.tag_attributes = tag_attributes

    def token_html(self, token_index):
        """The raw html for a page token at the given index in the page_tokens
        list
        """
        text_start = self.token_start_indexes[token_index]
        text_end = self.token_follow_indexes[token_index]
        return self.text[text_start:text_end]

    def html_between_tokens(self, start_token_index, end_token_index):
        """The raw html between the tokens at the specified indexes in the 
        page_tokens list

        This assumes start_token_index <= end_token_index
        """
        text_start = self.token_follow_indexes[start_token_index]
        text_end = self.token_start_indexes[end_token_index]
        return self.text[text_start:text_end]
    
    def text_between_tokens(self, start_token_index, end_token_index, 
            tag_replacement=u' '):
        """The text between the  the tokens at the specified indexes in the 
        page_tokens list. Tags are replaced by tag_replacement (default one space 
        character)
        """
        return tag_replacement.join([self.text[
            self.token_follow_indexes[i]:self.token_start_indexes[i+1]] \
            for i in xrange(start_token_index, end_token_index)])

    def tag_attribute(self, token_index, attribute):
        """The value of a tag attribute. The tag is identified by its 
        corresponding token index.

        If the tag or attribute is not present, None is returned
        """
        return self.tag_attributes.get(token_index, {}).get(attribute) 
    
    def __str__(self):
        summary = []
        for (token, start, follow) in zip(self.page_tokens, self.token_start_indexes, 
                self.token_follow_indexes):
            text = "%s %s-%s (%s)" % (self.token_dict.find_token(token), start, follow, 
                    self.text[start:follow])
            summary.append(text)
        return "ExtractionPage\n==============\nTokens: %s\n\nRaw text: %s\n\n" \
                "Tag attributes: %s\n" % ('\n'.join(summary), self.text, 
                        self.tag_attributes)

class AnnotationText(object):
    __slots__ = ('start_text', 'follow_text')

    def __init__(self, start_text=None, follow_text=None):
        self.start_text = start_text
        self.follow_text = follow_text

    def __str__(self):
        return "AnnotationText(%s..%s)" % \
                (repr(self.start_text), repr(self.follow_text))

class AnnotationTag(object):
    """A tag that annotates part of the document

    It has the following properties:
        start_index - index of the token for the opening tag
        end_index - index of the token for the closing tag
        surrounds_attribute - the attribute name surrounded by this tag
        tag_attributes - list of (tag attribute, extracted attribute) tuples
                         for each item to be extracted from a tag attribute
        annotation_text - text prefix and suffix for the attribute to be extracted
        match_common_prefix - use this annotation for calculating across-template prefixes
        metadata - dict with annotation data not used by IBL extractor
    """
    __slots__ = ('surrounds_attribute', 'start_index', 'end_index',
            'tag_attributes', 'annotation_text', 'variant_id', 
            'surrounds_variant','match_common_prefix', 'metadata')
    
    def __init__(self, start_index, end_index, surrounds_attribute=None, 
            annotation_text=None, tag_attributes=None, variant_id=None,
            surrounds_variant=None, match_common_prefix=False):
        self.start_index = start_index
        self.end_index = end_index
        self.surrounds_attribute = surrounds_attribute
        self.annotation_text = annotation_text
        self.tag_attributes = tag_attributes or []
        self.variant_id = variant_id
        self.surrounds_variant = surrounds_variant
        self.match_common_prefix = match_common_prefix
        self.metadata = {}

    def __str__(self):
        return "AnnotationTag(%s)" % ", ".join(
                ["%s=%s" % (s, getattr(self, s)) \
                for s in self.__slots__ if getattr(self, s)])

    def __repr__(self):
        return str(self)

class LabelledRegion(object):
    __slots__ = ('start_index', 'end_index')
    
    def __init__(self, start, end):
        self.start_index = start
        self.end_index = end
        
    def __str__(self):
        return "LabelledRegion (%s, %s)" % (self.start_index, self.end_index)
    
    def __repr__(self):
        return str(self)




from twisted.trial import unittest

from scrapy.http import Request

from scrapy.contrib_exp.crawlspider.reqproc import Canonicalize
from scrapy.contrib_exp.crawlspider.reqproc import FilterDomain
from scrapy.contrib_exp.crawlspider.reqproc import FilterUrl
from scrapy.contrib_exp.crawlspider.reqproc import FilterDupes

import copy

class RequestProcessorsTest(unittest.TestCase):

    def test_canonicalize_requests(self):
        urls = [
            'http://example.com/do?&b=1&a=2&c=3',
            'http://example.com/do?123,&q=a space',
            ]
        urls_after = [
            'http://example.com/do?a=2&b=1&c=3',
            'http://example.com/do?123%2C=&q=a+space',
            ]

        proc = Canonicalize()
        results = [req.url for req in proc(Request(url) for url in urls)]
        self.failUnlessEquals(results, urls_after)

    def test_unique_requests(self):
        urls = [
            'http://example.com/sample1.html',
            'http://example.com/sample2.html',
            'http://example.com/sample3.html',
            'http://example.com/sample1.html',
            'http://example.com/sample2.html',
            ]
        urls_unique = [
            'http://example.com/sample1.html',
            'http://example.com/sample2.html',
            'http://example.com/sample3.html',
            ]

        proc = FilterDupes()
        results = [req.url for req in proc(Request(url) for url in urls)]
        self.failUnlessEquals(results, urls_unique)

        # Check custom attributes
        requests = [
            Request('http://example.com', method='GET'),
            Request('http://example.com', method='POST'),
            ]
        proc = FilterDupes('url', 'method')
        self.failUnlessEqual(len(list(proc(requests))), 2)

        proc = FilterDupes('url')
        self.failUnlessEqual(len(list(proc(requests))), 1)

    def test_filter_domain(self):
        urls = [
            'http://blah1.com/index',
            'http://blah2.com/index',
            'http://blah1.com/section',
            'http://blah2.com/section',
            ]

        proc = FilterDomain(allow=('blah1.com'), deny=('blah2.com'))
        filtered = [req.url for req in proc(Request(url) for url in urls)]
        self.failUnlessEquals(filtered, [
                    'http://blah1.com/index',
                    'http://blah1.com/section',
                ])

        proc = FilterDomain(deny=('blah1.com', 'blah2.com'))
        filtered = [req.url for req in proc(Request(url) for url in urls)]
        self.failUnlessEquals(filtered, [])

        proc = FilterDomain(allow=('blah1.com', 'blah2.com'))
        filtered = [req.url for req in proc(Request(url) for url in urls)]
        self.failUnlessEquals(filtered, urls)

    def test_filter_url(self):
        urls = [
            'http://blah1.com/index',
            'http://blah2.com/index',
            'http://blah1.com/section',
            'http://blah2.com/section',
            ]

        proc = FilterUrl(allow=(r'blah1'), deny=(r'blah2'))
        filtered = [req.url for req in proc(Request(url) for url in urls)]
        self.failUnlessEquals(filtered, [
                    'http://blah1.com/index',
                    'http://blah1.com/section',
                ])

        proc = FilterUrl(deny=('blah1', 'blah2'))
        filtered = [req.url for req in proc(Request(url) for url in urls)]
        self.failUnlessEquals(filtered, [])

        proc = FilterUrl(allow=('index$', 'section$'))
        filtered = [req.url for req in proc(Request(url) for url in urls)]
        self.failUnlessEquals(filtered, urls)



    def test_all_processors(self):
        urls = [
            'http://example.com/sample1.html',
            'http://example.com/sample2.html',
            'http://example.com/sample3.html',
            'http://example.com/sample1.html',
            'http://example.com/sample2.html',
            'http://example.com/do?&b=1&a=2&c=3',
            'http://example.com/do?123,&q=a space',
            ]
        urls_processed = [
            'http://example.com/sample1.html',
            'http://example.com/sample2.html',
            'http://example.com/sample3.html',
            'http://example.com/do?a=2&b=1&c=3',
            'http://example.com/do?123%2C=&q=a+space',
            ]

        processors = [
            Canonicalize(),
            FilterDupes(),
            ]

        def _process(requests):
            """Apply all processors"""
            # copy list
            processed = [copy.copy(req) for req in requests]
            for proc in processors:
                processed = proc(processed)
            return processed

        # empty requests
        results1 = [r.url for r in _process([])]
        self.failUnlessEquals(results1, [])

        # try urls
        requests = (Request(url) for url in urls)
        results2 = [r.url for r in _process(requests)]
        self.failUnlessEquals(results2, urls_processed)




import sqlite3

from zope.interface import Interface, implements

from scrapy.xlib.pydispatch import dispatcher
from scrapy.exceptions import NotConfigured
from scrapy.utils.misc import load_object
from scrapy.utils.sqlite import JsonSqliteDict
from scrapy import log, signals

class ISpiderContextStorage(Interface):

    def get(spider):
        """Get the context for the given spider, or None if the spider has no
        context stored."""

    def put(spider, context):
        """Store the context for the given spider"""


class SqliteSpiderContextStorage(object):

    implements(ISpiderContextStorage)
    sqlite_dict_class = JsonSqliteDict

    def __init__(self, database=None, table='contexts'):
        try:
            self.d = self.sqlite_dict_class(database, table)
        except sqlite3.Error, e:
            self.d = self.sqlite_dict_class(':memory:', table)
            log.msg("Cannot open SQLite %r - using in-memory context storage " \
                "instead. Error was: %r" % (database, str(e)), log.WARNING)

    @classmethod
    def from_settings(cls, settings):
        return cls(settings['SQLITE_DB'])

    def get(self, spider):
        if spider.name in self.d:
            return self.d[spider.name]

    def put(self, spider, context):
        self.d[spider.name] = context


class SpiderContext(object):

    def __init__(self, storage):
        dispatcher.connect(self._spider_opened, signals.spider_opened)
        dispatcher.connect(self._spider_closed, signals.spider_closed)
        self.storage = storage

    @classmethod
    def from_settings(cls, settings):
        if not settings.getbool('SPIDER_CONTEXT_ENABLED'):
            raise NotConfigured
        stcls = load_object(settings['SPIDER_CONTEXT_STORAGE_CLASS'])
        storage = stcls.from_settings(settings)
        return cls(storage)

    def _spider_opened(self, spider):
        spider.context = self.storage.get(spider) or {}

    def _spider_closed(self, spider):
        if spider.context:
            self.storage.put(spider, spider.context)




"""Request Processors"""
from scrapy.utils.misc import arg_to_iter
from scrapy.utils.url import canonicalize_url, url_is_from_any_domain

from itertools import ifilter

import re

class Canonicalize(object):
    """Canonicalize Request Processor"""

    def __call__(self, requests):
        """Canonicalize all requests' urls"""
        return (x.replace(url=canonicalize_url(x.url)) for x in requests)
        

class FilterDupes(object):
    """Filter duplicate Requests"""

    def __init__(self, *attributes):
        """Initialize comparison attributes"""
        self._attributes = tuple(attributes) if attributes \
                                             else tuple(['url'])

    def _equal_attr(self, obj1, obj2, attr):
        return getattr(obj1, attr) == getattr(obj2, attr)

    def _requests_equal(self, req1, req2):
        """Attribute comparison helper"""
        # look for not equal attribute
        _not_equal = lambda attr: not self._equal_attr(req1, req2, attr)
        for attr in ifilter(_not_equal, self._attributes):
            return False
        # all attributes equal
        return True

    def _request_in(self, request, requests_seen):
        """Check if request is in given requests seen list"""
        _req_seen = lambda r: self._requests_equal(r, request)
        for seen in ifilter(_req_seen, requests_seen):
            return True
        # request not seen
        return False

    def __call__(self, requests):
        """Filter seen requests"""
        # per-call duplicates filter
        self.requests_seen = set()
        _not_seen = lambda r: not self._request_in(r, self.requests_seen)
        for req in ifilter(_not_seen, requests):
            yield req
            # registry seen request
            self.requests_seen.add(req)


class FilterDomain(object):
    """Filter request's domain"""

    def __init__(self, allow=(), deny=()):
         """Initialize allow/deny attributes"""
         self.allow = tuple(arg_to_iter(allow))
         self.deny = tuple(arg_to_iter(deny))

    def __call__(self, requests):
        """Filter domains"""
        processed = (req for req in requests)

        if self.allow:
            processed = (req for req in requests
                            if url_is_from_any_domain(req.url, self.allow))
        if self.deny:
            processed = (req for req in requests
                            if not url_is_from_any_domain(req.url, self.deny))

        return processed


class FilterUrl(object):
    """Filter request's url"""

    def __init__(self, allow=(), deny=()):
        """Initialize allow/deny attributes"""
        _re_type = type(re.compile('', 0))

        self.allow_res = [x if isinstance(x, _re_type) else re.compile(x) 
                          for x in arg_to_iter(allow)]
        self.deny_res = [x if isinstance(x, _re_type) else re.compile(x) 
                         for x in arg_to_iter(deny)]

    def __call__(self, requests):
        """Filter request's url based on allow/deny rules"""
        #TODO: filter valid urls here?
        processed = (req for req in requests)

        if self.allow_res:
            processed = (req for req in requests
                            if self._matches(req.url, self.allow_res))
        if self.deny_res:
            processed = (req for req in requests
                            if not self._matches(req.url, self.deny_res))

        return processed

    def _matches(self, url, regexs):
        """Returns True if url matches any regex in given list"""
        return any(r.search(url) for r in regexs)




import unittest

from scrapy.contrib.downloadermiddleware.downloadtimeout import DownloadTimeoutMiddleware
from scrapy.spider import BaseSpider
from scrapy.http import Request


class DownloadTimeoutMiddlewareTest(unittest.TestCase):

    def setUp(self):
        self.mw = DownloadTimeoutMiddleware()
        self.spider = BaseSpider('foo')
        self.req = Request('http://scrapytest.org/')

    def tearDown(self):
        del self.mw
        del self.spider
        del self.req

    def test_spider_has_no_download_timeout(self):
        assert self.mw.process_request(self.req, self.spider) is None
        assert 'download_timeout' not in self.req.meta

    def test_spider_has_download_timeout(self):
        self.spider.download_timeout = 2
        assert self.mw.process_request(self.req, self.spider) is None
        self.assertEquals(self.req.meta.get('download_timeout'), 2)

    def test_request_has_download_timeout(self):
        self.spider.download_timeout = 2
        self.req.meta['download_timeout'] = 1
        assert self.mw.process_request(self.req, self.spider) is None
        self.assertEquals(self.req.meta.get('download_timeout'), 1)
